"""
Pseudocode generation routines to create a Haxe representation of the decompiled IR.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, List

from .core import Bytecode, Type, Function, Fun
from . import disasm
from .decomp import (
    IRFunction,
    IRBlock,
    IRStatement,
    IRExpression,
    IRAssign,
    IRLocal,
    IRConst,
    IRArithmetic,
    IRBoolExpr,
    IRCall,
    IRConditional,
    IRWhileLoop,
    IRPrimitiveLoop,
    IRReturn,
    IRPrimitveJump,
    _get_type_in_code,
)


def _indent_str(level: int) -> str:
    return "    " * level  # 4 spaces for indentation


def _expression_to_haxe(expr: Optional[IRStatement], code: Bytecode, ir_function: IRFunction) -> str:
    assert expr is not None, "Found empty statement!"
    if isinstance(expr, IRLocal):
        return expr.name
    elif isinstance(expr, IRConst):
        if isinstance(expr.value, Function):  # crashlink.core.Function
            # For function constants, use their partial name or findex
            return disasm.partial_func_name(code, expr.value) or f"f@{expr.value.findex.value}"
        elif isinstance(expr.value, str):
            # Basic string quoting, may need more sophisticated escaping for real Haxe
            return '"' + expr.value.replace('"', '\\"') + '"'
        elif isinstance(expr.value, bool):
            return "true" if expr.value else "false"
        elif expr.value is None:  # For IRConst.ConstType.NULL
            return "null"
        return str(expr.value)
    elif isinstance(expr, IRArithmetic):
        left = _expression_to_haxe(expr.left, code, ir_function)
        right = _expression_to_haxe(expr.right, code, ir_function)
        # Add parentheses for potentially ambiguous operations if needed for clarity
        # e.g., if expr.op.value in ["*", "/"] and (isinstance(expr.left, IRArithmetic) or isinstance(expr.right, IRArithmetic)):
        # For pseudocode, direct representation is often fine.
        return f"{left} {expr.op.value} {right}"
    elif isinstance(expr, IRBoolExpr):
        op_map = {
            IRBoolExpr.CompareType.EQ: "==",
            IRBoolExpr.CompareType.NEQ: "!=",
            IRBoolExpr.CompareType.LT: "<",
            IRBoolExpr.CompareType.LTE: "<=",
            IRBoolExpr.CompareType.GT: ">",
            IRBoolExpr.CompareType.GTE: ">=",
        }
        if expr.op == IRBoolExpr.CompareType.NULL:
            return f"{_expression_to_haxe(expr.left, code, ir_function)} == null"
        elif expr.op == IRBoolExpr.CompareType.NOT_NULL:
            return f"{_expression_to_haxe(expr.left, code, ir_function)} != null"
        elif expr.op == IRBoolExpr.CompareType.ISTRUE:
            # For Haxe, (expr) or (expr == true)
            return _expression_to_haxe(expr.left, code, ir_function)
        elif expr.op == IRBoolExpr.CompareType.ISFALSE:
            return f"!{_expression_to_haxe(expr.left, code, ir_function)}"
        elif expr.op == IRBoolExpr.CompareType.NOT:
            return f"!{_expression_to_haxe(expr.left, code, ir_function)}"
        elif expr.op == IRBoolExpr.CompareType.TRUE:
            return "true"
        elif expr.op == IRBoolExpr.CompareType.FALSE:
            return "false"
        elif expr.left and expr.right and expr.op in op_map:
            left = _expression_to_haxe(expr.left, code, ir_function)
            right = _expression_to_haxe(expr.right, code, ir_function)
            return f"{left} {op_map[expr.op]} {right}"
        elif expr.left:  # Unary boolean expressions like NOT handled above
            raise NotImplementedError(f"Unhandled unary IRBoolExpr op: {expr.op} on {expr.left}")
        else:
            raise NotImplementedError(f"Unhandled IRBoolExpr: {expr}")

    elif isinstance(expr, IRCall):
        callee_str: str
        if expr.call_type == IRCall.CallType.THIS and expr.target is None:
            # This assumes the method name is somehow retrievable or you have a convention
            # For now, let's assume a placeholder if the method name isn't directly in IRCall
            # You might need to pass the Opcode field for 'CallThis' to get the field name.
            callee_str = "this.unknownMethod"  # Placeholder
        elif expr.target:
            callee_str = _expression_to_haxe(expr.target, code, ir_function)
        else:  # Should have a target or be THIS
            raise ValueError(f"IRCall missing target or unhandled type: {expr.call_type}")

        args_str = ", ".join(_expression_to_haxe(arg, code, ir_function) for arg in expr.args)
        return f"{callee_str}({args_str})"
    elif isinstance(expr, IRPrimitveJump):  # Should be gone, but as a fallback
        return f"/* GOTO_LIKE({expr.op.op}) */"

    # Fallback for unknown expressions
    return f"/* <UnknownExpr: {type(expr).__name__}> */"


def _generate_statements(
    statements: List[IRStatement],
    code: Bytecode,
    ir_function: IRFunction,
    indent_level: int,
    # Track declared variables in the current scope to decide between "var x =" and "x ="
    # This is a simplification; a proper symbol table would be more robust.
    declared_vars_in_scope: set[str],
) -> List[str]:
    output_lines: List[str] = []
    indent = _indent_str(indent_level)

    for stmt in statements:
        if isinstance(stmt, IRBlock):  # Nested block, usually from if/else/loop bodies
            # HaxeBlock's content is generated by recursively calling _generate_statements
            # The parent (if/while) handles the "{" and "}"
            output_lines.extend(
                _generate_statements(stmt.statements, code, ir_function, indent_level, declared_vars_in_scope.copy())
            )
        elif isinstance(stmt, IRAssign):
            target_str = _expression_to_haxe(stmt.target, code, ir_function)
            value_str = _expression_to_haxe(stmt.expr, code, ir_function)

            # Simple check for declaration: if target is an IRLocal and not yet declared
            if isinstance(stmt.target, IRLocal) and stmt.target.name not in declared_vars_in_scope:
                type_name = disasm.type_to_haxe(disasm.type_name(code, stmt.target.get_type()))
                type_decl = f":{type_name}" if type_name and type_name != "Dynamic" and type_name != "Void" else ""
                output_lines.append(f"{indent}var {target_str}{type_decl} = {value_str};")
                declared_vars_in_scope.add(stmt.target.name)
            else:
                output_lines.append(f"{indent}{target_str} = {value_str};")

        elif isinstance(stmt, IRConditional):
            cond_str = _expression_to_haxe(stmt.condition, code, ir_function)
            output_lines.append(f"{indent}if ({cond_str}) {{")
            output_lines.extend(
                _generate_statements(
                    stmt.true_block.statements, code, ir_function, indent_level + 1, declared_vars_in_scope.copy()
                )
            )
            if stmt.false_block and stmt.false_block.statements:
                output_lines.append(f"{indent}}} else {{")
                output_lines.extend(
                    _generate_statements(
                        stmt.false_block.statements, code, ir_function, indent_level + 1, declared_vars_in_scope.copy()
                    )
                )
                output_lines.append(f"{indent}}}")
            else:
                output_lines.append(f"{indent}}}")

        elif isinstance(stmt, IRWhileLoop):
            cond_str = _expression_to_haxe(stmt.condition, code, ir_function)
            output_lines.append(f"{indent}while ({cond_str}) {{")
            output_lines.extend(
                _generate_statements(
                    stmt.body.statements, code, ir_function, indent_level + 1, declared_vars_in_scope.copy()
                )
            )
            output_lines.append(f"{indent}}}")

        elif isinstance(stmt, IRPrimitiveLoop):  # Fallback if not optimized to IRWhileLoop
            output_lines.append(f"{indent}// Primitive Loop (condition first, then body)")
            output_lines.append(f"{indent}{{ // Condition Block")
            output_lines.extend(
                _generate_statements(
                    stmt.condition.statements, code, ir_function, indent_level + 1, declared_vars_in_scope.copy()
                )
            )
            output_lines.append(f"{indent}}}")
            output_lines.append(f"{indent}{{ // Body Block")
            output_lines.extend(
                _generate_statements(
                    stmt.body.statements, code, ir_function, indent_level + 1, declared_vars_in_scope.copy()
                )
            )
            output_lines.append(f"{indent}}}")

        elif isinstance(stmt, IRReturn):
            if stmt.value:
                if isinstance(stmt.value, IRLocal) and stmt.value.type.resolve(code).kind.value == Type.Kind.VOID.value:
                    output_lines.append(
                        f"{indent}return; // implicit void return from reg{ir_function.locals.index(stmt.value) + 1}"
                    )
                else:
                    val_str = _expression_to_haxe(stmt.value, code, ir_function)
                    output_lines.append(f"{indent}return {val_str};")
            else:
                output_lines.append(f"{indent}return;")

        elif isinstance(stmt, IRExpression):  # e.g. a standalone IRCall not assigned
            expr_str = _expression_to_haxe(stmt, code, ir_function)
            output_lines.append(f"{indent}{expr_str};")

        else:
            output_lines.append(f"{indent}// <Unhandled IRStatement: {type(stmt).__name__}> {str(stmt)[:50]}...")

        if stmt.comment:
            # Add comment at the end of the line or on a new line
            if output_lines:
                output_lines[-1] += f" // {stmt.comment}"
            else:  # Should not happen if statement generated something
                output_lines.append(f"{indent}// {stmt.comment}")

    return output_lines


def pseudo(ir_func: IRFunction) -> str:
    """
    Generates Haxe pseudocode from a given IRFunction.
    """
    code: Bytecode = ir_func.code
    func_core: Function = ir_func.func  # crashlink.core.Function

    output_lines: List[str] = []
    base_indent = 0

    # Function Signature
    func_name_str = disasm.partial_func_name(code, func_core) or f"f{func_core.findex.value}"
    static_kw = "static " if disasm.is_static(code, func_core) else ""

    params_str_list = []
    return_type_str = "Void"  # Default

    core_fun_type_def = func_core.type.resolve(code).definition
    if isinstance(core_fun_type_def, Fun):  # crashlink.core.Fun
        for i, arg_type_idx in enumerate(core_fun_type_def.args):
            arg_core_type = arg_type_idx.resolve(code)
            arg_haxe_type_name = disasm.type_to_haxe(disasm.type_name(code, arg_core_type))

            param_name = f"arg{i}"  # Default name
            # Try to get actual param name from debug assigns (op index < 0)
            if func_core.has_debug and func_core.assigns:
                arg_assigns = [a for a in func_core.assigns if a[1].value < 0]
                if i < len(arg_assigns):
                    param_name = arg_assigns[i][0].resolve(code)

            param_type_decl = f":{arg_haxe_type_name}" if arg_haxe_type_name and arg_haxe_type_name != "Dynamic" else ""
            params_str_list.append(f"{param_name}{param_type_decl}")

        ret_core_type = core_fun_type_def.ret.resolve(code)
        return_type_str = disasm.type_to_haxe(disasm.type_name(code, ret_core_type))

    params_joined_str = ", ".join(params_str_list)
    func_header = (
        f"{_indent_str(base_indent)}{static_kw}function {func_name_str}({params_joined_str}):{return_type_str} {{"
    )
    output_lines.append(func_header)

    # Function Body
    # Initialize declared_vars with function parameters
    initial_declared_vars = {p_name.split(":")[0] for p_name in params_str_list}

    body_lines = _generate_statements(ir_func.block.statements, code, ir_func, base_indent + 1, initial_declared_vars)
    output_lines.extend(body_lines)

    output_lines.append(f"{_indent_str(base_indent)}}}")

    # Attempt to wrap in a class for context if class name can be derived
    full_name = disasm.full_func_name(code, func_core)
    class_name_suggestion = "DecompiledClass"
    if "." in full_name:
        class_name_part = full_name.split(".")[0]
        if class_name_part and class_name_part != "<none>":
            class_name_suggestion = class_name_part.replace("$", "_")  # Handle Haxe internal names

    final_output = [f"class {class_name_suggestion} {{"]
    final_output.extend(["  " + line for line in output_lines])  # Indent function within class
    final_output.append("}")

    return "\n".join(final_output)



"""
Definitions of all 98 supported opcodes in the HashLink VM.
"""

conditionals = [
    "JTrue",
    "JFalse",
    "JNull",
    "JNotNull",
    "JSLt",
    "JSGte",
    "JSGt",
    "JSLte",
    "JULt",
    "JUGte",
    "JNotLt",
    "JNotGte",
    "JEq",
    "JNotEq",
]
"""
List of all conditional opcodes.
"""

arithmetic = [
    "Add",
    "Sub",
    "Mul",
    "SDiv",
    "UDiv",
    "SMod",
    "UMod",
    "Shl",
    "SShr",
    "UShr",
    "And",
    "Or",
    "Xor",
]
"""
List of all arithmetic opcodes.
"""

simple_calls = ["Call0", "Call1", "Call2", "Call3", "Call4", "CallN"]
"""
List of all function call opcodes.
"""

opcodes = {
    "Mov": {"dst": "Reg", "src": "Reg"},  # 0
    "Int": {"dst": "Reg", "ptr": "RefInt"},  # 1
    "Float": {"dst": "Reg", "ptr": "RefFloat"},  # 2
    "Bool": {"dst": "Reg", "value": "InlineBool"},  # 3
    "Bytes": {"dst": "Reg", "ptr": "RefBytes"},  # 4
    "String": {"dst": "Reg", "ptr": "RefString"},  # 5
    "Null": {"dst": "Reg"},  # 6
    "Add": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 7
    "Sub": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 8
    "Mul": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 9
    "SDiv": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 10
    "UDiv": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 11
    "SMod": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 12
    "UMod": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 13
    "Shl": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 14
    "SShr": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 15
    "UShr": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 16
    "And": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 17
    "Or": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 18
    "Xor": {"dst": "Reg", "a": "Reg", "b": "Reg"},  # 19
    "Neg": {"dst": "Reg", "src": "Reg"},  # 20
    "Not": {"dst": "Reg", "src": "Reg"},  # 21
    "Incr": {"dst": "Reg"},  # 22
    "Decr": {"dst": "Reg"},  # 23
    "Call0": {"dst": "Reg", "fun": "RefFun"},  # 24
    "Call1": {"dst": "Reg", "fun": "RefFun", "arg0": "Reg"},  # 25
    "Call2": {"dst": "Reg", "fun": "RefFun", "arg0": "Reg", "arg1": "Reg"},  # 26
    "Call3": {
        "dst": "Reg",
        "fun": "RefFun",
        "arg0": "Reg",
        "arg1": "Reg",
        "arg2": "Reg",
    },  # 27
    "Call4": {
        "dst": "Reg",
        "fun": "RefFun",
        "arg0": "Reg",
        "arg1": "Reg",
        "arg2": "Reg",
        "arg3": "Reg",
    },  # 28
    "CallN": {"dst": "Reg", "fun": "RefFun", "args": "Regs"},  # 29
    "CallMethod": {"dst": "Reg", "field": "RefField", "args": "Regs"},  # 30
    "CallThis": {"dst": "Reg", "field": "RefField", "args": "Regs"},  # 31
    "CallClosure": {"dst": "Reg", "fun": "Reg", "args": "Regs"},  # 32
    "StaticClosure": {"dst": "Reg", "fun": "RefFun"},  # 33
    "InstanceClosure": {"dst": "Reg", "fun": "RefFun", "obj": "Reg"},  # 34
    "VirtualClosure": {"dst": "Reg", "obj": "Reg", "field": "Reg"},  # 35
    "GetGlobal": {"dst": "Reg", "global": "RefGlobal"},  # 36
    "SetGlobal": {"global": "RefGlobal", "src": "Reg"},  # 37
    "Field": {"dst": "Reg", "obj": "Reg", "field": "RefField"},  # 38
    "SetField": {"obj": "Reg", "field": "RefField", "src": "Reg"},  # 39
    "GetThis": {"dst": "Reg", "field": "RefField"},  # 40
    "SetThis": {"field": "RefField", "src": "Reg"},  # 41
    "DynGet": {"dst": "Reg", "obj": "Reg", "field": "RefString"},  # 42
    "DynSet": {"obj": "Reg", "field": "RefString", "src": "Reg"},  # 43
    "JTrue": {"cond": "Reg", "offset": "JumpOffset"},  # 44
    "JFalse": {"cond": "Reg", "offset": "JumpOffset"},  # 45
    "JNull": {"reg": "Reg", "offset": "JumpOffset"},  # 46
    "JNotNull": {"reg": "Reg", "offset": "JumpOffset"},  # 47
    "JSLt": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 48
    "JSGte": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 49
    "JSGt": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 50
    "JSLte": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 51
    "JULt": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 52
    "JUGte": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 53
    "JNotLt": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 54
    "JNotGte": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 55
    "JEq": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 56
    "JNotEq": {"a": "Reg", "b": "Reg", "offset": "JumpOffset"},  # 57
    "JAlways": {"offset": "JumpOffset"},  # 58
    "ToDyn": {"dst": "Reg", "src": "Reg"},  # 59
    "ToSFloat": {"dst": "Reg", "src": "Reg"},  # 60
    "ToUFloat": {"dst": "Reg", "src": "Reg"},  # 61
    "ToInt": {"dst": "Reg", "src": "Reg"},  # 62
    "SafeCast": {"dst": "Reg", "src": "Reg"},  # 63
    "UnsafeCast": {"dst": "Reg", "src": "Reg"},  # 64
    "ToVirtual": {"dst": "Reg", "src": "Reg"},  # 65
    "Label": {},  # 66
    "Ret": {"ret": "Reg"},  # 67
    "Throw": {"exc": "Reg"},  # 68
    "Rethrow": {"exc": "Reg"},  # 69
    "Switch": {"reg": "Reg", "offsets": "JumpOffsets", "end": "JumpOffset"},  # 70
    "NullCheck": {"reg": "Reg"},  # 71
    "Trap": {"exc": "Reg", "offset": "JumpOffset"},  # 72
    "EndTrap": {"exc": "Reg"},  # 73
    "GetI8": {"dst": "Reg", "bytes": "Reg", "index": "Reg"},  # 74
    "GetI16": {"dst": "Reg", "bytes": "Reg", "index": "Reg"},  # 75
    "GetMem": {"dst": "Reg", "bytes": "Reg", "index": "Reg"},  # 76
    "GetArray": {"dst": "Reg", "array": "Reg", "index": "Reg"},  # 77
    "SetI8": {"bytes": "Reg", "index": "Reg", "src": "Reg"},  # 78
    "SetI16": {"bytes": "Reg", "index": "Reg", "src": "Reg"},  # 79
    "SetMem": {"bytes": "Reg", "index": "Reg", "src": "Reg"},  # 80
    "SetArray": {"array": "Reg", "index": "Reg", "src": "Reg"},  # 81
    "New": {"dst": "Reg"},  # 82
    "ArraySize": {"dst": "Reg", "array": "Reg"},  # 83
    "Type": {"dst": "Reg", "ty": "RefType"},  # 84
    "GetType": {"dst": "Reg", "src": "Reg"},  # 85
    "GetTID": {"dst": "Reg", "src": "Reg"},  # 86
    "Ref": {"dst": "Reg", "src": "Reg"},  # 87
    "Unref": {"dst": "Reg", "src": "Reg"},  # 88
    "Setref": {"dst": "Reg", "value": "Reg"},  # 89
    "MakeEnum": {"dst": "Reg", "construct": "RefEnumConstruct", "args": "Regs"},  # 90
    "EnumAlloc": {"dst": "Reg", "construct": "RefEnumConstruct"},  # 91
    "EnumIndex": {"dst": "Reg", "value": "Reg"},  # 92
    "EnumField": {
        "dst": "Reg",
        "value": "Reg",
        "construct": "RefEnumConstruct",
        "field": "RefField",
    },  # 93
    "SetEnumField": {"value": "Reg", "field": "RefField", "src": "Reg"},  # 94
    "Assert": {},  # 95
    "RefData": {"dst": "Reg", "src": "Reg"},  # 96
    "RefOffset": {"dst": "Reg", "reg": "Reg", "offset": "Reg"},  # 97
    "Nop": {},  # 98
    "Prefetch": {"value": "Reg", "field": "RefField", "mode": "InlineInt"},  # 99
    "Asm": {"mode": "InlineInt", "value": "InlineInt", "reg": "Reg"},  # 100
}
"""
Definitions of all 98 supported opcodes in the HashLink VM.

Basic Operations:

- Mov: Copy value from src into dst register (`dst = src`) 
- Int: Load i32 constant from pool into dst (`dst = @ptr`)
- Float: Load f64 constant from pool into dst (`dst = @ptr`)
- Bool: Set boolean value in dst (`dst = true/false`)
- Bytes: Load byte array from constant pool into dst (`dst = @ptr`)
- String: Load string from constant pool into dst (`dst = @ptr`)
- Null: Set dst register to null (`dst = null`)

Arithmetic:

- Add: Add two numbers (`dst = a + b`)
- Sub: Subtract two numbers (`dst = a - b`) 
- Mul: Multiply two numbers (`dst = a * b`)
- SDiv: Signed division (`dst = a / b`)
- UDiv: Unsigned division (`dst = a / b`)
- SMod: Signed modulo (`dst = a % b`)
- UMod: Unsigned modulo (`dst = a % b`)

Bitwise:

- Shl: Left shift (`dst = a << b`)
- SShr: Signed right shift (`dst = a >> b`)
- UShr: Unsigned right shift (`dst = a >>> b`)
- And: Bitwise AND (`dst = a & b`)
- Or: Bitwise OR (`dst = a | b`)
- Xor: Bitwise XOR (`dst = a ^ b`)
- Neg: Negate value (`dst = -src`)
- Not: Boolean NOT (`dst = !src`)

Increment/Decrement:

- Incr: Increment value (`dst++`)
- Decr: Decrement value (`dst--`)

Function Calls:

- Call0: Call function with no args (`dst = fun()`)
- Call1: Call function with 1 arg (`dst = fun(arg0)`)
- Call2: Call function with 2 args (`dst = fun(arg0, arg1)`)
- Call3: Call function with 3 args (`dst = fun(arg0, arg1, arg2)`)
- Call4: Call function with 4 args (`dst = fun(arg0, arg1, arg2, arg3)`)
- CallN: Call function with N args (`dst = fun(args...)`)
- CallMethod: Call method with N args (`dst = obj.field(args...)`)
- CallThis: Call this method with N args (`dst = this.field(args...)`)
- CallClosure: Call closure with N args (`dst = fun(args...)`)

Closures:

- StaticClosure: Create closure from function (`dst = fun`)
- InstanceClosure: Create closure from object method (`dst = obj.fun`)
- VirtualClosure: Create closure from object field (`dst = obj.field`)

Global Variables:

- GetGlobal: Get global value (`dst = @global`)
- SetGlobal: Set global value (`@global = src`)

Fields:

- Field: Get object field (`dst = obj.field`)  
- SetField: Set object field (`obj.field = src`)
- GetThis: Get this field (`dst = this.field`)
- SetThis: Set this field (`this.field = src`)
- DynGet: Get dynamic field (`dst = obj[field]`)
- DynSet: Set dynamic field (`obj[field] = src`)

Control Flow:

- JTrue: Jump if true (`if cond jump by offset`)
- JFalse: Jump if false (`if !cond jump by offset`)
- JNull: Jump if null (`if reg == null jump by offset`)
- JNotNull: Jump if not null (`if reg != null jump by offset`)
- JSLt/JSGte/JSGt/JSLte: Signed comparison jumps
- JULt/JUGte: Unsigned comparison jumps 
- JNotLt/JNotGte: Negated comparison jumps
- JEq: Jump if equal (`if a == b jump by offset`)
- JNotEq: Jump if not equal (`if a != b jump by offset`)
- JAlways: Unconditional jump
- Label: Target for backward jumps (loops)
- Switch: Multi-way branch based on integer value

Type Conversions:

- ToDyn: Convert to dynamic type (`dst = (dyn)src`)
- ToSFloat: Convert to signed float (`dst = (float)src`)
- ToUFloat: Convert to unsigned float (`dst = (float)src`) 
- ToInt: Convert to int (`dst = (int)src`)
- SafeCast: Safe type cast with runtime check
- UnsafeCast: Unchecked type cast
- ToVirtual: Convert to virtual type

Exception Handling:

- Ret: Return from function (`return ret`)
- Throw: Throw exception 
- Rethrow: Rethrow exception
- Trap: Setup try-catch block
- EndTrap: End try-catch block
- NullCheck: Throw if null (`if reg == null throw`)

Memory Operations:

- GetI8: Read i8 from bytes (`dst = bytes[index]`)
- GetI16: Read i16 from bytes (`dst = bytes[index]`)
- GetMem: Read from memory (`dst = bytes[index]`)
- GetArray: Get array element (`dst = array[index]`)
- SetI8: Write i8 to bytes (`bytes[index] = src`)
- SetI16: Write i16 to bytes (`bytes[index] = src`) 
- SetMem: Write to memory (`bytes[index] = src`)
- SetArray: Set array element (`array[index] = src`)

Objects:

- New: Allocate new object (`dst = new typeof(dst)`)
- ArraySize: Get array length (`dst = len(array)`)
- Type: Get type object (`dst = type ty`)
- GetType: Get value's type (`dst = typeof src`)
- GetTID: Get type ID (`dst = typeof src`)

References:

- Ref: Create reference (`dst = &src`)
- Unref: Read reference (`dst = *src`)
- Setref: Write reference (`*dst = src`)
- RefData: Get reference data
- RefOffset: Get reference with offset

Enums:

- MakeEnum: Create enum variant (`dst = construct(args...)`)
- EnumAlloc: Create enum with defaults (`dst = construct()`)
- EnumIndex: Get enum tag (`dst = variant of value`)
- EnumField: Get enum field (`dst = (value as construct).field`)
- SetEnumField: Set enum field (`value.field = src`)

Other:

- Assert: Debug break
- Nop: No operation
- Prefetch: CPU memory prefetch hint
- Asm: Inline x86 assembly

If you want to see a more detailed pseudocode for any given instance of `Opcode`, you can use `crashlink.disasm.pseudo_from_op()` to get a human-readable representation of the operation.

If you're using the CLI in patch mode, you'll see opcodes in a format like:

```hlasm
Int. 0. 0
Int. 2. 1
GetGlobal. 3. 3
Add. 4. 0. 2
Sub. 5. 0. 2
Mul. 6. 0. 2
ToSFloat. 8. 0
ToSFloat. 9. 2
SDiv. 8. 8. 9
SMod. 7. 0. 2
Shl. 10. 0. 2
JSLt. 0. 2. 2
Bool. 11. False
JAlways. 1
Bool. 11. True
JSLt. 0. 2. 2
Bool. 12. False
JAlways. 1
Bool. 12. True
Nop.
CallN. [1, 2, 3]
Ret. 1
```

When writing opcodes in this format, separate each argument to the opcode with a period. For opcodes that require lists, pass them in JSON format, e.g. `CallN. [reg, reg, reg, reg]`.
"""

opcode_docs = {
    # Basic Operations
    "Mov": "Copy value from src into dst register (dst = src)",
    "Int": "Load i32 constant from pool into dst (dst = @ptr)",
    "Float": "Load f64 constant from pool into dst (dst = @ptr)",
    "Bool": "Set boolean value in dst (dst = true/false)",
    "Bytes": "Load byte array from constant pool into dst (dst = @ptr)",
    "String": "Load string from constant pool into dst (dst = @ptr)",
    "Null": "Set dst register to null (dst = null)",
    # Arithmetic
    "Add": "Add two numbers (dst = a + b)",
    "Sub": "Subtract two numbers (dst = a - b)",
    "Mul": "Multiply two numbers (dst = a * b)",
    "SDiv": "Signed division (dst = a / b)",
    "UDiv": "Unsigned division (dst = a / b)",
    "SMod": "Signed modulo (dst = a % b)",
    "UMod": "Unsigned modulo (dst = a % b)",
    # Bitwise
    "Shl": "Left shift (dst = a << b)",
    "SShr": "Signed right shift (dst = a >> b)",
    "UShr": "Unsigned right shift (dst = a >>> b)",
    "And": "Bitwise AND (dst = a & b)",
    "Or": "Bitwise OR (dst = a | b)",
    "Xor": "Bitwise XOR (dst = a ^ b)",
    "Neg": "Negate value (dst = -src)",
    "Not": "Boolean NOT (dst = !src)",
    # Increment/Decrement
    "Incr": "Increment value (dst++)",
    "Decr": "Decrement value (dst--)",
    # Function Calls
    "Call0": "Call function with no args (dst = fun())",
    "Call1": "Call function with 1 arg (dst = fun(arg0))",
    "Call2": "Call function with 2 args (dst = fun(arg0, arg1))",
    "Call3": "Call function with 3 args (dst = fun(arg0, arg1, arg2))",
    "Call4": "Call function with 4 args (dst = fun(arg0, arg1, arg2, arg3))",
    "CallN": "Call function with N args (dst = fun(args...))",
    "CallMethod": "Call method with N args (dst = obj.field(args...))",
    "CallThis": "Call this method with N args (dst = this.field(args...))",
    "CallClosure": "Call closure with N args (dst = fun(args...))",
    # Closures
    "StaticClosure": "Create closure from function (dst = fun)",
    "InstanceClosure": "Create closure from object method (dst = obj.fun)",
    "VirtualClosure": "Create closure from object field (dst = obj.field)",
    # Global Variables
    "GetGlobal": "Get global value (dst = @global)",
    "SetGlobal": "Set global value (@global = src)",
    # Fields
    "Field": "Get object field (dst = obj.field)",
    "SetField": "Set object field (obj.field = src)",
    "GetThis": "Get this field (dst = this.field)",
    "SetThis": "Set this field (this.field = src)",
    "DynGet": "Get dynamic field (dst = obj[field])",
    "DynSet": "Set dynamic field (obj[field] = src)",
    # Control Flow
    "JTrue": "Jump if true (if cond jump by offset)",
    "JFalse": "Jump if false (if !cond jump by offset)",
    "JNull": "Jump if null (if reg == null jump by offset)",
    "JNotNull": "Jump if not null (if reg != null jump by offset)",
    "JSLt": "Signed comparison jump (less than)",
    "JSGte": "Signed comparison jump (greater than or equal)",
    "JSGt": "Signed comparison jump (greater than)",
    "JSLte": "Signed comparison jump (less than or equal)",
    "JULt": "Unsigned comparison jump (less than)",
    "JUGte": "Unsigned comparison jump (greater than or equal)",
    "JNotLt": "Negated comparison jump (not less than)",
    "JNotGte": "Negated comparison jump (not greater than or equal)",
    "JEq": "Jump if equal (if a == b jump by offset)",
    "JNotEq": "Jump if not equal (if a != b jump by offset)",
    "JAlways": "Unconditional jump",
    "Label": "Target for backward jumps (loops)",
    "Switch": "Multi-way branch based on integer value",
    # Type Conversions
    "ToDyn": "Convert to dynamic type (dst = (dyn)src)",
    "ToSFloat": "Convert to signed float (dst = (float)src)",
    "ToUFloat": "Convert to unsigned float (dst = (float)src)",
    "ToInt": "Convert to int (dst = (int)src)",
    "SafeCast": "Safe type cast with runtime check",
    "UnsafeCast": "Unchecked type cast",
    "ToVirtual": "Convert to virtual type",
    # Exception Handling
    "Ret": "Return from function (return ret)",
    "Throw": "Throw exception",
    "Rethrow": "Rethrow exception",
    "Trap": "Setup try-catch block",
    "EndTrap": "End try-catch block",
    "NullCheck": "Throw if null (if reg == null throw)",
    # Memory Operations
    "GetI8": "Read i8 from bytes (dst = bytes[index])",
    "GetI16": "Read i16 from bytes (dst = bytes[index])",
    "GetMem": "Read from memory (dst = bytes[index])",
    "GetArray": "Get array element (dst = array[index])",
    "SetI8": "Write i8 to bytes (bytes[index] = src)",
    "SetI16": "Write i16 to bytes (bytes[index] = src)",
    "SetMem": "Write to memory (bytes[index] = src)",
    "SetArray": "Set array element (array[index] = src)",
    # Objects
    "New": "Allocate new object (dst = new typeof(dst))",
    "ArraySize": "Get array length (dst = len(array))",
    "Type": "Get type object (dst = type ty)",
    "GetType": "Get value's type (dst = typeof src)",
    "GetTID": "Get type ID (dst = typeof src)",
    # References
    "Ref": "Create reference (dst = &src)",
    "Unref": "Read reference (dst = *src)",
    "Setref": "Write reference (*dst = src)",
    "RefData": "Get reference data",
    "RefOffset": "Get reference with offset",
    # Enums
    "MakeEnum": "Create enum variant (dst = construct(args...))",
    "EnumAlloc": "Create enum with defaults (dst = construct())",
    "EnumIndex": "Get enum tag (dst = variant of value)",
    "EnumField": "Get enum field (dst = (value as construct).field)",
    "SetEnumField": "Set enum field (value.field = src)",
    # Other
    "Assert": "Debug break",
    "Nop": "No operation",
    "Prefetch": "CPU memory prefetch hint",
    "Asm": "Inline x86 assembly",
}

__all__ = ["conditionals", "arithmetic", "opcodes", "opcode_docs"]



from typing import Tuple

from crashlink import *


def load_test(path) -> Tuple[Bytecode, Function]:
    code = Bytecode.from_path(path)
    assert code is not None
    func = None
    for f in code.functions:
        if "main" in disasm.full_func_name(code, f):
            func = f
            break
    assert func is not None
    return code, func


def test_switch():
    code, func = load_test("tests/haxe/Switch.hl")
    cfg = decomp.CFGraph(func)
    cfg.build()
    assert cfg.nodes[0].ops[-1].op == "Switch"
    assert cfg.nodes[-1].ops[-1].op == "Ret"



"""
Global configuration.
"""

from __future__ import annotations

import os
from io import BytesIO
from typing import Any, BinaryIO


def _is_debug() -> bool:
    """
    Determine whether or not to enable debug mode.
    """
    return (
        bool(os.getenv("CRASHLINK_DEBUG", False))
        or bool(os.getenv("DEBUG", False))
        or os.path.exists(".crashlink_debug")
    )


VERSION: str = "v0.0.4"
"""
The version of crashlink.
"""

LONG_VERSION: str = "crashlink - Pure-Python HashLink bytecode Swiss Army knife - " + VERSION
"""
String displayed in the help message for the CLI.
"""

DEBUG: bool = False or _is_debug()
"""
Whether to enable certain features meant only for development or debugging of crashlink.
"""


def dbg_print(*args: Any, **kwargs: Any) -> None:
    """
    Print a message if DEBUG is True.
    """
    if DEBUG:
        print(*args, **kwargs)


def tell(f: BinaryIO | BytesIO) -> str:
    """
    Hex-formatted tell of a file.
    """
    return hex(f.tell())


def fmt_bytes(bytes: int | float) -> str:
    """
    Format bytes into a human-readable string.
    """
    if bytes < 0:
        raise ValueError("Bytes cannot be negative.")

    size_units = ["B", "Kb", "Mb", "Gb", "Tb"]
    index = 0

    while bytes >= 1000 and index < len(size_units) - 1:
        bytes /= 1000
        index += 1

    return f"{bytes:.1f}{size_units[index]}"


# https://stackoverflow.com/a/287944
class bcolors:
    """
    ANSI color escape sequences.
    """

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


__all__ = [
    "VERSION",
    "LONG_VERSION",
    "DEBUG",
    "dbg_print",
    "tell",
    "fmt_bytes",
    "bcolors",
]



"""
Runtime tools for instrumenting Hashlink bytecode. See `crashlink.patch`.

WARNING: This module should never be used alone - you should use it in conjunction with pyhl, or, better yet, just use `crashlink.patch`.
"""

from .core import *
from .globals import *
from .obj import *



"""
Core bytecode format definitions.

This module contains the definitions for HashLink bytecode structures, as well as the serialisation
and deserialisation methods for them. You probably don't need to use too much of this file directly,
besides Bytecode, Opcode, and Function. The decompiler will take care of a lot of abstraction for
you.
"""

from __future__ import annotations

import ctypes
import struct
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum as _Enum
from io import BytesIO
from typing import Any, BinaryIO, Dict, List, Literal, Optional, Tuple, TypeVar

T = TypeVar("T", bound="VarInt")  # easier than reimplementing deserialise for each subclass

from .errors import InvalidOpCode, MalformedBytecode, NoMagic
from .globals import dbg_print, tell
from .opcodes import opcodes, simple_calls

try:
    from tqdm import tqdm

    USE_TQDM = True
except ImportError:
    dbg_print("Could not find tqdm. Progress bars will not be displayed.")
    USE_TQDM = False


class Serialisable(ABC):
    """
    Base class for all serialisable objects.
    """

    @abstractmethod
    def __init__(self) -> None:
        self.value: Any = None

    @abstractmethod
    def deserialise(self, f: BinaryIO | BytesIO, *args: Any, **kwargs: Any) -> "Serialisable":
        pass

    @abstractmethod
    def serialise(self) -> bytes:
        pass

    def __str__(self) -> str:
        try:
            return str(self.value)
        except AttributeError:
            return super().__repr__()

    def __repr__(self) -> str:
        try:
            return str(self.value)
        except AttributeError:
            return super().__repr__()

    def __eq__(self, other: object) -> Any:
        if not isinstance(other, Serialisable):
            return NotImplemented
        return self.value == other.value

    def __ne__(self, other: object) -> Any:
        if not isinstance(other, Serialisable):
            return NotImplemented
        return self.value != other.value

    def __lt__(self, other: object) -> Any:
        if not isinstance(other, Serialisable):
            return NotImplemented
        return self.value < other.value


class RawData(Serialisable):
    """
    A block of raw data.
    """

    def __init__(self, length: int):
        self.value: bytes = b""
        self.length = length

    def deserialise(self, f: BinaryIO | BytesIO) -> "RawData":
        self.value = f.read(self.length)
        return self

    def serialise(self) -> bytes:
        return self.value


class SerialisableInt(Serialisable):
    """
    Integer of the specified byte length.
    """

    def __init__(self) -> None:
        self.value: int = -1
        self.length = 4
        self.byteorder: Literal["little", "big"] = "little"
        self.signed = False

    def deserialise(
        self,
        f: BinaryIO | BytesIO,
        length: int = 4,
        byteorder: Literal["little", "big"] = "little",
        signed: bool = False,
    ) -> "SerialisableInt":
        self.length = length
        self.byteorder = byteorder
        self.signed = signed
        bytes = f.read(length)
        if all(b == 0 for b in bytes):
            self.value = 0
            return self
        while bytes[-1] == 0:
            bytes = bytes[:-1]
        self.value = int.from_bytes(bytes, byteorder, signed=signed)
        return self

    def serialise(self) -> bytes:
        return self.value.to_bytes(self.length, self.byteorder, signed=self.signed)


class SerialisableF64(Serialisable):
    """
    A standard 64-bit float.
    """

    def __init__(self) -> None:
        self.value = 0.0

    def deserialise(self, f: BinaryIO | BytesIO) -> "SerialisableF64":
        self.value = struct.unpack("<d", f.read(8))[0]
        return self

    def serialise(self) -> bytes:
        return struct.pack("<d", self.value)


_struct_short = struct.Struct(">H")  # big-endian unsigned short
_struct_medium = struct.Struct(">I")  # big-endian unsigned int (for 3 bytes)


class VarInt(Serialisable):
    """
    Variable-length integer - can be 1, 2, or 4 bytes.
    """

    def __init__(self, value: int = 0):
        self.value: int = value

    def deserialise(self: T, f: BinaryIO | BytesIO) -> T:
        # Read first byte - keep int.from_bytes for single byte
        b = int.from_bytes(f.read(1), "big")

        # Single byte format (0xxxxxxx)
        if not (b & 0x80):
            self.value = b
            return self

        # Two byte format (10xxxxxx)
        if not (b & 0x40):
            # Read 2 bytes as unsigned short
            second = f.read(1)[0]  # Faster than int.from_bytes for single byte

            # Combine bytes and handle sign
            self.value = ((b & 0x1F) << 8) | second
            if b & 0x20:
                self.value = -self.value
            return self

        # Four byte format (11xxxxxx)
        remaining = _struct_medium.unpack(b"\x00" + f.read(3))[0]

        # Combine all bytes and handle sign
        self.value = ((b & 0x1F) << 24) | remaining
        if b & 0x20:
            self.value = -self.value
        return self

    def serialise(self) -> bytes:
        if self.value < 0:
            value = -self.value
            if value < 0x2000:  # 13 bits
                return bytes([(value >> 8) | 0xA0, value & 0xFF])
            if value >= 0x20000000:
                raise MalformedBytecode("value can't be >= 0x20000000")
            # Optimized 4-byte case
            return bytes(
                [
                    (value >> 24) | 0xE0,
                    (value >> 16) & 0xFF,
                    (value >> 8) & 0xFF,
                    value & 0xFF,
                ]
            )

        if self.value < 0x80:  # 7 bits
            return bytes([self.value])
        if self.value < 0x2000:  # 13 bits
            return bytes([(self.value >> 8) | 0x80, self.value & 0xFF])
        if self.value >= 0x20000000:
            raise MalformedBytecode("value can't be >= 0x20000000")
        # Optimized 4-byte case
        return bytes(
            [
                (self.value >> 24) | 0xC0,
                (self.value >> 16) & 0xFF,
                (self.value >> 8) & 0xFF,
                self.value & 0xFF,
            ]
        )


class ResolvableVarInt(VarInt, ABC):
    """
    Base class for resolvable VarInts. Call `resolve` to get a direct reference to the object it points to.
    """

    @abstractmethod
    def resolve(self, code: "Bytecode") -> Any:
        """
        Resolve this reference to a specific reference in the bytecode.
        """
        pass


class fIndex(ResolvableVarInt):
    """
    Abstract class based on VarInt to represent a distinct function index instead of just an arbitrary number.
    """

    def resolve(self, code: "Bytecode") -> "Function|Native":
        for function in code.functions:
            if function.findex.value == self.value:
                return function
        for native in code.natives:
            if native.findex.value == self.value:
                return native
        raise MalformedBytecode(f"Function index {self.value} not found.")


class tIndex(ResolvableVarInt):
    """
    Reference to a type in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> "Type":
        return code.types[self.value]


class gIndex(ResolvableVarInt):
    """
    Reference to a global object in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> "Type":
        return code.global_types[self.value].resolve(code)

    def resolve_str(self, code: "Bytecode") -> str:
        return code.const_str(self.value)


class strRef(ResolvableVarInt):
    """
    Reference to a string in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> str:
        return code.strings.value[self.value]


class intRef(ResolvableVarInt):
    """
    Reference to an integer in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> SerialisableInt:
        return code.ints[self.value]


class floatRef(ResolvableVarInt):
    """
    Reference to a float in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> SerialisableF64:
        return code.floats[self.value]


class bytesRef(ResolvableVarInt):
    """
    Reference to a byte string in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> bytes:
        if code.bytes:
            return code.bytes.value[self.value]
        else:
            raise MalformedBytecode("No bytes block found.")


class fieldRef(ResolvableVarInt):
    """
    Reference to a field in an object definition.
    """

    obj: Optional["Obj|Virtual"] = None

    def resolve(self, code: "Bytecode") -> "Field":
        if self.obj:
            return self.obj.resolve_fields(code)[self.value]
        raise ValueError(
            "Cannot resolve field without context. Try setting `field.obj` to an instance of `Obj`, or use `field.resolve_obj(code, obj)` instead."
        )

    def resolve_obj(self, code: "Bytecode", obj: "Obj|Virtual") -> "Field":
        self.obj = obj
        return obj.resolve_fields(code)[self.value]


class Reg(ResolvableVarInt):
    """
    Reference to a register in the bytecode.
    """

    def resolve(self, code: "Bytecode") -> "Type":
        return code.types[self.value]


class InlineBool(Serialisable):
    """
    Inline boolean value.
    """

    def __init__(self) -> None:
        self.varint = VarInt()
        self.value: bool = False

    def deserialise(self, f: BinaryIO | BytesIO) -> "InlineBool":
        self.varint.deserialise(f)
        self.value = bool(self.varint.value)
        return self

    def serialise(self) -> bytes:
        self.varint.value = int(self.value)
        return self.varint.serialise()


class VarInts(Serialisable):
    """
    List of VarInts.
    """

    def __init__(self) -> None:
        self.n = VarInt()
        self.value: List[VarInt] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "VarInts":
        self.n.deserialise(f)
        for _ in range(self.n.value):
            self.value.append(VarInt().deserialise(f))
        return self

    def serialise(self) -> bytes:
        return b"".join([self.n.serialise(), b"".join([value.serialise() for value in self.value])])


class Regs(Serialisable):
    """
    List of references to registers.
    """

    def __init__(self) -> None:
        self.n = VarInt()
        self.value: List[Reg] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "Regs":
        self.n.deserialise(f)
        for _ in range(self.n.value):
            self.value.append(Reg().deserialise(f))
        return self

    def serialise(self) -> bytes:
        return b"".join([self.n.serialise(), b"".join([value.serialise() for value in self.value])])


class StringsBlock(Serialisable):
    """
    Block of strings in the bytecode. Contains a list of strings and their lengths.
    """

    def __init__(self) -> None:
        self.length = SerialisableInt()
        self.length.length = 4
        self.value: List[str] = []
        self.lengths: List[VarInt] = []

    def deserialise(self, f: BinaryIO | BytesIO, nstrings: int) -> "StringsBlock":
        self.length.deserialise(f, length=4)
        size = self.length.value
        sdata: bytes = f.read(size)
        strings: List[str] = []
        lengths: List[VarInt] = []
        curpos = 0

        for _ in range(nstrings):
            sz = VarInt().deserialise(f)
            # Check if we can read string + null terminator
            if curpos + sz.value + 1 > size:
                raise ValueError("Invalid string")

            # Verify null terminator
            if sdata[curpos + sz.value] != 0:
                raise ValueError("Invalid string")

            str_value = sdata[curpos : curpos + sz.value]
            strings.append(str_value.decode("utf-8", errors="surrogateescape"))
            lengths.append(sz)

            curpos += sz.value + 1  # Move past string and null terminator

        self.value = strings
        self.lengths = lengths
        return self

    def serialise(self) -> bytes:
        strings_data = bytearray()
        for string in self.value:
            encoded = string.encode("utf-8", errors="surrogateescape")
            strings_data.extend(encoded)
            strings_data.append(0)  # null terminator

        self.length.value = len(strings_data)
        self.lengths = [VarInt(len(string.encode("utf-8", errors="surrogateescape"))) for string in self.value]

        result = bytearray(self.length.serialise())
        result.extend(strings_data)
        for length in self.lengths:
            result.extend(length.serialise())

        return bytes(result)

    def find_or_add(self, val: str) -> int:
        """
        Finds and returns the index of a string value in this block, or adds it to this block and returns its index.
        """
        if val in self.value:
            return self.value.index(val)
        self.value.append(val)
        return len(self.value) - 1


class BytesBlock(Serialisable):
    """
    Block of bytes in the bytecode. Contains a list of byte strings and their lengths.
    """

    def __init__(self) -> None:
        self.size = SerialisableInt()
        self.size.length = 4
        self.value: List[bytes] = []
        self.nbytes = 0

    def deserialise(self, f: BinaryIO | BytesIO, nbytes: int) -> "BytesBlock":
        self.nbytes = nbytes
        self.size.deserialise(f, length=4)
        raw = f.read(self.size.value)
        positions: List[VarInt] = []
        for _ in range(nbytes):
            pos = VarInt()
            pos.deserialise(f)
            positions.append(pos)
        positions_int = [pos.value for pos in positions]
        for i in range(len(positions_int)):
            start = positions_int[i]
            end = positions_int[i + 1] if i + 1 < len(positions_int) else len(raw)
            self.value.append(raw[start:end])  # Append the extracted byte string
        return self

    def serialise(self) -> bytes:
        raw_data = b"".join(self.value)
        self.size.value = len(raw_data)
        size_serialised = self.size.serialise()
        positions = []
        current_pos = 0
        for byte_str in self.value:
            positions.append(VarInt(current_pos))
            current_pos += len(byte_str)
        positions_serialised = b"".join([pos.serialise() for pos in positions])
        return size_serialised + raw_data + positions_serialised


class TypeDef(Serialisable):
    """
    Abstract class for all type definition fields.
    """


class _NoDataType(TypeDef):
    """
    Base typedef for types with no data.
    """

    def __init__(self) -> None:
        pass

    def deserialise(self, f: BinaryIO | BytesIO) -> "_NoDataType":
        return self

    def serialise(self) -> bytes:
        return b""


class Void(_NoDataType):
    """
    Void type, no data. Used to discard data (eg. reg: Void = call f@**).
    """

    pass


class U8(_NoDataType):
    """
    Unsigned 8-bit integer type, no data.
    """

    pass


class U16(_NoDataType):
    """
    Unsigned 16-bit integer type, no data.
    """

    pass


class I32(_NoDataType):
    """
    Signed 32-bit integer type, no data.
    """

    pass


class I64(_NoDataType):
    """
    Signed 64-bit integer type, no data.
    """

    pass


class F32(_NoDataType):
    """
    32-bit float type, no data.
    """

    pass


class F64(_NoDataType):
    """
    64-bit float type, no data.
    """

    pass


class Bool(_NoDataType):
    """
    Boolean type, no data.
    """

    pass


class Bytes(_NoDataType):
    """
    Bytes type, no data.
    """

    pass


class Dyn(_NoDataType):
    """
    Dynamic type, no data. Can store any type of data in a typed register as a pointer.
    """

    pass


class Fun(TypeDef):
    """
    Stores metadata about a function (signatures). When referenced in conjunction with a Proto or Method, it can be used to reconstruct the full function signature. See `crashlink.disasm.func_header` for a working reference.
    """

    def __init__(self) -> None:
        self.nargs = VarInt()
        self.args: List[tIndex] = []
        self.ret = tIndex()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Fun":
        self.nargs.deserialise(f)
        for _ in range(self.nargs.value):
            self.args.append(tIndex().deserialise(f))
        self.ret.deserialise(f)
        return self

    def serialise(self) -> bytes:
        self.nargs.value = len(self.args)
        return b"".join(
            [
                self.nargs.serialise(),
                b"".join([idx.serialise() for idx in self.args]),
                self.ret.serialise(),
            ]
        )


class Field(Serialisable):
    """
    Represents a field in a class definition.
    """

    def __init__(self, name: Optional[strRef] = None, type: Optional[tIndex] = None) -> None:
        if not name:
            self.name = strRef()
        else:
            self.name = name
        if not type:
            self.type = tIndex()
        else:
            self.type = type

    def deserialise(self, f: BinaryIO | BytesIO) -> "Field":
        self.name.deserialise(f)
        self.type.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return b"".join([self.name.serialise(), self.type.serialise()])


class Proto(Serialisable):
    """
    Represents a prototype of a function
    """

    def __init__(self) -> None:
        self.name = strRef()
        self.findex = fIndex()
        self.pindex = VarInt()  # unknown use

    def deserialise(self, f: BinaryIO | BytesIO) -> "Proto":
        self.name.deserialise(f)
        self.findex.deserialise(f)
        self.pindex.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return b"".join([self.name.serialise(), self.findex.serialise(), self.pindex.serialise()])


class Binding(Serialisable):
    """
    Represents a binding of a field to a class.
    """

    def __init__(self) -> None:
        self.field = fieldRef()
        self.findex = fIndex()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Binding":
        self.field.deserialise(f)
        self.findex.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return b"".join([self.field.serialise(), self.findex.serialise()])


class Obj(TypeDef):
    """
    Represents a class definition. Contains:

    - name: strRef
    - super: tIndex
    - _global: gIndex
    - nfields: VarInt
    - nprotos: VarInt
    - nbindings: VarInt
    - fields: List[Field]
    - protos: List[Proto]
    - bindings: List[Binding]
    """

    def __init__(self) -> None:
        self.name = strRef()
        self.super = tIndex()
        self._global = gIndex()
        self.nfields = VarInt()
        self.nprotos = VarInt()
        self.nbindings = VarInt()
        self.fields: List[Field] = []
        self.protos: List[Proto] = []
        self.bindings: List[Binding] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "Obj":
        self.name.deserialise(f)
        self.super.deserialise(f)
        self._global.deserialise(f)
        self.nfields.deserialise(f)
        self.nprotos.deserialise(f)
        self.nbindings.deserialise(f)
        for _ in range(self.nfields.value):
            self.fields.append(Field().deserialise(f))
        for _ in range(self.nprotos.value):
            self.protos.append(Proto().deserialise(f))
        for _ in range(self.nbindings.value):
            self.bindings.append(Binding().deserialise(f))
        return self

    def serialise(self) -> bytes:
        return b"".join(
            [
                self.name.serialise(),
                self.super.serialise(),
                self._global.serialise(),
                self.nfields.serialise(),
                self.nprotos.serialise(),
                self.nbindings.serialise(),
                b"".join([field.serialise() for field in self.fields]),
                b"".join([proto.serialise() for proto in self.protos]),
                b"".join([binding.serialise() for binding in self.bindings]),
            ]
        )

    def resolve_fields(self, code: "Bytecode") -> List[Field]:
        """
        Resolves all fields across the class heirarchy. For instance:
        class A {
            var a: Int;
        }
        class B extends A {
            var b: Int;
        }
        Where a is field 0 and b is field 1
        """
        if self.super.value < 0:  # no superclass
            return self.fields
        fields: List[Field] = []
        visited_types = set()
        current_type: Optional[Obj] = self
        while current_type:
            if id(current_type) in visited_types:
                raise ValueError("Cyclic inheritance detected in class hierarchy.")
            visited_types.add(id(current_type))
            fields = current_type.fields + fields
            if current_type.super.value < 0:
                current_type = None
            else:
                defn = current_type.super.resolve(code).definition
                if not isinstance(defn, Obj):
                    raise ValueError("Invalid superclass type.")
                current_type = defn
        return fields

    def __str__(self) -> str:
        return f"<Obj: f@{self.name}>"

    def __repr__(self) -> str:
        return self.__str__()

    def str_resolve(self, code: "Bytecode") -> str:
        return f"<Obj: {self.name.resolve(code)}>"


class Array(_NoDataType):
    """
    Array type, no data.
    """

    pass


class TypeType(_NoDataType):
    """
    Type wrapping a type, no data.
    """

    pass


class Ref(TypeDef):
    """
    Memory reference to an instance of a type.
    """

    def __init__(self) -> None:
        self.type = tIndex()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Ref":
        self.type.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return self.type.serialise()


class Virtual(TypeDef):
    """
    Virtual type, used for virtual/abstract classes.
    """

    def __init__(self) -> None:
        self.nfields = VarInt()
        self.fields: List[Field] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "Virtual":
        self.nfields.deserialise(f)
        for _ in range(self.nfields.value):
            self.fields.append(Field().deserialise(f))
        return self

    def serialise(self) -> bytes:
        self.nfields.value = len(self.fields)
        return b"".join(
            [
                self.nfields.serialise(),
                b"".join([field.serialise() for field in self.fields]),
            ]
        )

    def resolve_fields(self, code: "Bytecode") -> List[Field]:
        return self.fields


class DynObj(_NoDataType):
    """
    Dynamic object type, no data.
    """

    pass


class Abstract(TypeDef):
    """
    Abstract class type.
    """

    def __init__(self) -> None:
        self.name = strRef()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Abstract":
        self.name.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return self.name.serialise()


class EnumConstruct(Serialisable):
    """
    Construct of an enum.
    """

    def __init__(self) -> None:
        self.name = strRef()
        self.nparams = VarInt()
        self.params: List[tIndex] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "EnumConstruct":
        self.name.deserialise(f)
        self.nparams.deserialise(f)
        for _ in range(self.nparams.value):
            self.params.append(tIndex().deserialise(f))
        return self

    def serialise(self) -> bytes:
        return b"".join(
            [
                self.name.serialise(),
                self.nparams.serialise(),
                b"".join([param.serialise() for param in self.params]),
            ]
        )


class Enum(TypeDef):
    """
    Enum type.
    """

    def __init__(self) -> None:
        self.name = strRef()
        self._global = gIndex()
        self.nconstructs = VarInt()
        self.constructs: List[EnumConstruct] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "Enum":
        self.name.deserialise(f)
        self._global.deserialise(f)
        self.nconstructs.deserialise(f)
        for _ in range(self.nconstructs.value):
            self.constructs.append(EnumConstruct().deserialise(f))
        return self

    def serialise(self) -> bytes:
        return b"".join(
            [
                self.name.serialise(),
                self._global.serialise(),
                self.nconstructs.serialise(),
                b"".join([construct.serialise() for construct in self.constructs]),
            ]
        )


class Null(TypeDef):
    """
    Null of a certain type.
    """

    def __init__(self) -> None:
        self.type = tIndex()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Null":
        self.type.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return self.type.serialise()


class Method(Fun):
    """
    Method type, identical to Fun.
    """

    pass


class Struct(Obj):
    """
    Struct type, identical to Obj.
    """

    pass


class Packed(TypeDef):
    """
    Holds an inner type index.
    """

    def __init__(self) -> None:
        self.inner = tIndex()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Packed":
        self.inner.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return self.inner.serialise()


class Type(Serialisable):
    """
    Type definition:

    - kind: SerialisableInt
    - definition: TypeDef
    """

    # fmt: off
    TYPEDEFS: List[type] = [
        Void,     # 0, no data
        U8,       # 1, no data
        U16,      # 2, no data
        I32,      # 3, no data
        I64,      # 4, no data
        F32,      # 5, no data
        F64,      # 6, no data
        Bool,     # 7, no data
        Bytes,    # 8, no data
        Dyn,      # 9, no data
        Fun,      # 10
        Obj,      # 11
        Array,    # 12, no data
        TypeType, # 13, no data
        Ref,      # 14
        Virtual,  # 15
        DynObj,   # 16, no data
        Abstract, # 17
        Enum,     # 18
        Null,     # 19
        Method,   # 20
        Struct,   # 21
        Packed,   # 22
    ]
    # fmt: on

    class Kind(_Enum):
        VOID = 0
        U8 = 1
        U16 = 2
        I32 = 3
        I64 = 4
        F32 = 5
        F64 = 6
        BOOL = 7
        BYTES = 8
        DYN = 9
        FUN = 10
        OBJ = 11
        ARRAY = 12
        TYPETYPE = 13
        REF = 14
        VIRTUAL = 15
        DYNOBJ = 16
        ABSTRACT = 17
        ENUM = 18
        NULL = 19
        METHOD = 20
        STRUCT = 21
        PACKED = 22

    def __init__(self) -> None:
        self.kind = SerialisableInt()
        self.kind.length = 1
        self.definition: Optional[TypeDef] = None

    def deserialise(self, f: BinaryIO | BytesIO) -> "Type":
        # dbg_print(f"Type @ {tell(f)}")
        self.kind.deserialise(f, length=1)
        try:
            self.TYPEDEFS[self.kind.value]
            _def = self.TYPEDEFS[self.kind.value]()
            self.definition = _def.deserialise(f)
        except IndexError:
            raise MalformedBytecode(f"Invalid type kind found @{tell(f)}")
        return self

    def serialise(self) -> bytes:
        return b"".join(
            [
                self.kind.serialise(),
                self.definition.serialise() if self.definition else b"",
            ]
        )

    def __str__(self) -> str:
        return f"<Type: {self.kind.value} ({self.definition.__class__.__name__})>"

    def __repr__(self) -> str:
        return self.__str__()

    def str_resolve(self, code: "Bytecode") -> str:
        if isinstance(self.definition, Obj):
            return self.definition.str_resolve(code)
        return f"<Type: {self.kind.value} ({self.definition.__class__.__name__})>"


class Native(Serialisable):
    """
    Represents a native function.

    - lib: strRef
    - name: strRef
    - type: tIndex
    - findex: fIndex
    """

    def __init__(self) -> None:
        self.lib = strRef()
        self.name = strRef()
        self.type = tIndex()
        self.findex = fIndex()

    def deserialise(self, f: BinaryIO | BytesIO) -> "Native":
        self.lib.deserialise(f)
        self.name.deserialise(f)
        self.type.deserialise(f)
        self.findex.deserialise(f)
        return self

    def serialise(self) -> bytes:
        return b"".join(
            [
                self.lib.serialise(),
                self.name.serialise(),
                self.type.serialise(),
                self.findex.serialise(),
            ]
        )


class Opcode(Serialisable):
    """
    Represents an opcode.
    """

    TYPE_MAP: Dict[str, type] = {
        "Reg": Reg,
        "Regs": Regs,
        "RefInt": intRef,
        "RefFloat": floatRef,
        "InlineBool": InlineBool,
        "RefBytes": bytesRef,
        "RefString": strRef,
        "RefFun": fIndex,
        "RefField": fieldRef,
        "RefGlobal": gIndex,
        "JumpOffset": VarInt,
        "JumpOffsets": VarInts,
        "RefType": tIndex,
        "RefEnumConstant": VarInt,
        "RefEnumConstruct": VarInt,
        "InlineInt": VarInt,
    }

    def __init__(self, op: Optional[str] = None, df: Optional[Dict[Any, Any]] = None) -> None:
        self.code = VarInt()
        self.op: Optional[str] = None
        if op:
            self.op = op
        self.df: Dict[Any, Any] = {}
        if df:
            self.df = df

    def deserialise(self, f: BinaryIO | BytesIO) -> "Opcode":
        # dbg_print(f"Deserialising opcode at {tell(f)}... ", end="")
        self.code.deserialise(f)
        # dbg_print(f"{self.code.value}... ", end="")
        try:
            _def = opcodes[list(opcodes.keys())[self.code.value]]
        except IndexError:
            raise InvalidOpCode(f"Unknown opcode at {tell(f)} - {self.code.value}")
        for param, _type in _def.items():
            if _type in self.TYPE_MAP:
                self.df[param] = self.TYPE_MAP[_type]().deserialise(f)
                continue
            raise InvalidOpCode(f"Invalid opcode definition for {param, _type} at {tell(f)}")
        self.op = list(opcodes.keys())[self.code.value]
        return self

    def serialise(self) -> bytes:
        if self.op:
            self.code.value = list(opcodes.keys()).index(self.op)
        return b"".join(
            [
                self.code.serialise(),
                b"".join([definition.serialise() for name, definition in self.df.items()]),
            ]
        )

    def __repr__(self) -> str:
        return f"<Opcode: {self.op} {self.df}>"

    def __str__(self) -> str:
        return self.__repr__()


class fileRef(ResolvableVarInt):
    """
    Reference to a file in the debug info.
    """

    def __init__(self, fid: int = 0, line: int = -1) -> None:
        super().__init__(fid)
        self.line = line

    def resolve(self, code: "Bytecode") -> str:
        """
        Resolve to the filename of the reference.
        """
        if not code.debugfiles:
            raise MalformedBytecode("No debug files found.")
        return code.debugfiles.value[self.value]

    def resolve_line(self, code: "Bytecode") -> int:
        """
        Resolve to the line number in the file of the reference.
        """
        return self.line

    def resolve_pretty(self, code: "Bytecode") -> str:
        """
        Resolve a pretty-printed string: <filename>:<line number>
        """
        return f"{self.resolve(code)}:{self.line}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, fileRef):
            return NotImplemented
        return self.value == other.value and self.line == other.line


class DebugInfo(Serialisable):
    """
    Represents debug information for a function, encoded with a delta encoding scheme for compression.
    """

    def __init__(self) -> None:
        self.value: List[fileRef] = []

    def deserialise(self, f: BinaryIO | BytesIO, nops: int) -> "DebugInfo":
        tmp = []
        currfile: int = -1
        currline: int = 0
        i = 0
        while i < nops:
            c = ctypes.c_uint8(ord(f.read(1))).value
            if c & 1 != 0:
                c >>= 1
                currfile = (c << 8) | ctypes.c_uint8(ord(f.read(1))).value
            elif c & 2 != 0:
                delta = c >> 6
                count = (c >> 2) & 15
                while count > 0:
                    count -= 1
                    tmp.append(fileRef(currfile, currline))
                    i += 1
                currline += delta
            elif c & 4 != 0:
                currline += c >> 3
                tmp.append(fileRef(currfile, currline))
                i += 1
            else:
                b2 = ctypes.c_uint8(ord(f.read(1))).value
                b3 = ctypes.c_uint8(ord(f.read(1))).value
                currline = (c >> 3) | (b2 << 5) | (b3 << 13)
                tmp.append(fileRef(currfile, currline))
                i += 1
        self.value = tmp
        return self

    def _flush_repeat(
        self,
        w: BinaryIO | BytesIO,
        curpos: ctypes.c_size_t,
        rcount: ctypes.c_size_t,
        pos: int,
    ) -> None:
        if rcount.value > 0:
            if rcount.value > 15:
                w.write(ctypes.c_uint8((15 << 2) | 2).value.to_bytes(1, "little"))
                rcount.value -= 15
                self._flush_repeat(w, curpos, rcount, pos)
            else:
                delta = pos - curpos.value
                delta = delta if 0 < delta < 4 else 0
                w.write(ctypes.c_uint8(((delta << 6) | (rcount.value << 2) | 2)).value.to_bytes(1, "little"))
                rcount.value = 0
                curpos.value += delta

    def serialise(self) -> bytes:
        w = BytesIO()
        curfile = -1
        curpos = ctypes.c_size_t(0)
        rcount = ctypes.c_size_t(0)

        for ref in self.value:
            f = ref.value
            p = ref.line
            if f != curfile:
                self._flush_repeat(w, curpos, rcount, p)
                curfile = f
                w.write(ctypes.c_uint8(((f >> 7) | 1)).value.to_bytes(1, "little"))
                w.write(ctypes.c_uint8(f & 0xFF).value.to_bytes(1, "little"))

            if p != curpos.value:
                self._flush_repeat(w, curpos, rcount, p)

            if p == curpos.value:
                rcount.value += 1
            else:
                delta = p - curpos.value
                if 0 < delta < 32:
                    w.write(ctypes.c_uint8((delta << 3) | 4).value.to_bytes(1, "little"))
                else:
                    w.write(ctypes.c_uint8((p << 3) & 0xFF).value.to_bytes(1, "little"))
                    w.write(ctypes.c_uint8((p >> 5) & 0xFF).value.to_bytes(1, "little"))
                    w.write(ctypes.c_uint8((p >> 13) & 0xFF).value.to_bytes(1, "little"))
                curpos.value = p

        self._flush_repeat(w, curpos, rcount, curpos.value)

        return w.getvalue()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DebugInfo):
            return NotImplemented
        return self.value == other.value


class Function(Serialisable):
    """
    Represents a function in the bytecode. Due to the interesting ways in which HashLink works, this does not have a name or a signature, but rather a return type and a list of registers and opcodes.
    """

    def __init__(self) -> None:
        self.type = tIndex()
        self.findex = fIndex()
        self.nregs = VarInt()
        self.nops = VarInt()
        self.regs: List[tIndex] = []
        self.ops: List[Opcode] = []
        self.has_debug: Optional[bool] = None
        self.version: Optional[int] = None
        self.debuginfo: Optional[DebugInfo] = None
        self.nassigns: Optional[VarInt] = None
        self.assigns: Optional[List[Tuple[strRef, VarInt]]] = None
        self.calls: List[fIndex] = []

    def called_by(self, code: "Bytecode") -> List[fIndex]:
        """
        Resolves all functions that call this function.
        """
        caller_indices = []
        for func in code.functions:
            if any(call_idx.value == self.findex.value for call_idx in func.calls):
                caller_indices.append(func.findex)
        return caller_indices

    def resolve_fun(self, code: "Bytecode") -> Fun:
        """
        Resolves the function signature of this function.
        """
        ret = self.type.resolve(code).definition
        assert ret is not None
        assert isinstance(ret, Fun)
        return ret

    def resolve_file(self, code: "Bytecode") -> str:
        """
        Resolves (in the Bytecode's debugfiles blob) the name of the file this function originates from. Note that this assumes the first opcode's file is the only file this Function was derived from - Functions that derive from multiple files (such as compiler-generated closures or entrypoints) will be resolved to a single file, sometimes incorrectly.
        """
        if not self.has_debug or not self.debuginfo:
            raise ValueError("Cannot get file from non-debug bytecode!")
        return self.debuginfo.value[0].resolve(code)

    def deserialise(self, f: BinaryIO | BytesIO, has_debug: bool, version: int) -> "Function":
        self.has_debug = has_debug
        self.version = version
        self.type.deserialise(f)
        self.findex.deserialise(f)
        self.nregs.deserialise(f)
        self.nops.deserialise(f)
        for _ in range(self.nregs.value):
            self.regs.append(tIndex().deserialise(f))
        for _ in range(self.nops.value):
            self.ops.append(Opcode().deserialise(f))
            if self.ops[-1].op in simple_calls:
                self.calls.append(self.ops[-1].df["fun"])
        if self.has_debug:
            self.debuginfo = DebugInfo().deserialise(f, self.nops.value)
            if self.version >= 3:
                self.nassigns = VarInt().deserialise(f)
                self.assigns = []
                for _ in range(self.nassigns.value):
                    self.assigns.append((strRef().deserialise(f), VarInt().deserialise(f)))
        return self

    def insert_op(self, code: "Bytecode", idx: int, op: Opcode, debugRef: Optional[fileRef] = None) -> None:
        """
        Insert an Opcode into this function at the given position, adding a blank debug fileRef if none is passed.
        """
        self.ops.insert(idx, op)
        if code.debugfiles and code.has_debug_info and self.has_debug and self.debuginfo:  # fucking typing...
            if not debugRef:
                debugRef = fileRef(fid=code.debugfiles.find_or_add("?"), line=42)  # life, the universe, and everything
            self.debuginfo.value.insert(idx, debugRef)

    def push_op(self, code: "Bytecode", op: Opcode, debugRef: Optional[fileRef] = None) -> int:
        """
        Push an Opcode into this function at the start.
        """
        self.insert_op(code, 0, op, debugRef=debugRef)
        return 0

    def append_op(self, code: "Bytecode", op: Opcode, debugRef: Optional[fileRef] = None) -> int:
        """
        Append an Opcode to the end of this function.
        """
        self.insert_op(code, -1, op, debugRef=debugRef)
        return len(self.ops) - 1

    def serialise(self) -> bytes:
        self.nops.value = len(self.ops)
        self.nregs.value = len(self.regs)
        if self.assigns:
            self.nassigns = VarInt(len(self.assigns) if self.assigns else 0)
        if self.has_debug and self.debuginfo:
            assert len(self.debuginfo.value) == self.nops.value, (
                f"Invalid number of debugrefs - {len(self.debuginfo.value)} (debuginfo) != {self.nops.value} (nops) - did you use insert_op?"
            )
        res = b"".join(
            [
                self.type.serialise(),
                self.findex.serialise(),
                self.nregs.serialise(),
                self.nops.serialise(),
                b"".join([reg.serialise() for reg in self.regs]),
                b"".join([op.serialise() for op in self.ops]),
            ]
        )
        if self.has_debug and self.debuginfo:
            res += self.debuginfo.serialise()
            if self.version and self.version >= 3 and self.nassigns and self.assigns is not None:
                res += self.nassigns.serialise()
                res += b"".join([b"".join([v.serialise() for v in assign]) for assign in self.assigns])
        return res


class Constant(Serialisable):
    """
    Represents a bytecode constant.
    """

    def __init__(self) -> None:
        self._global = gIndex()
        self.nfields = VarInt()
        self.fields: List[VarInt] = []

    def deserialise(self, f: BinaryIO | BytesIO) -> "Constant":
        self._global.deserialise(f)
        self.nfields.deserialise(f)
        for _ in range(self.nfields.value):
            self.fields.append(VarInt().deserialise(f))
        return self

    def serialise(self) -> bytes:
        return b"".join(
            [
                self._global.serialise(),
                self.nfields.serialise(),
                b"".join([field.serialise() for field in self.fields]),
            ]
        )


class Bytecode(Serialisable):
    """
    The main bytecode class. To read a bytecode file, use the `from_path` class method.

    For more information about the overall structure, see [here](https://n3rdl0rd.github.io/ModDocCE/files/hlboot)
    """

    def __init__(self) -> None:
        self.deserialised = False
        self.magic = RawData(3)
        self.version = SerialisableInt()
        self.version.length = 1
        self.flags = VarInt()
        self.has_debug_info: Optional[bool] = None
        self.nints = VarInt()
        self.nfloats = VarInt()
        self.nstrings = VarInt()
        self.nbytes: Optional[VarInt] = VarInt()
        self.ntypes = VarInt()
        self.nglobals = VarInt()
        self.nnatives = VarInt()
        self.nfunctions = VarInt()
        self.nconstants: Optional[VarInt] = VarInt()
        self.entrypoint = fIndex()

        self.ints: List[SerialisableInt] = []
        self.floats: List[SerialisableF64] = []
        self.strings = StringsBlock()
        self.bytes: Optional[BytesBlock] = BytesBlock()

        self.ndebugfiles: Optional[VarInt] = VarInt()
        self.debugfiles: Optional[StringsBlock] = StringsBlock()

        self.types: List[Type] = []
        self.global_types: List[tIndex] = []
        self.natives: List[Native] = []
        self.functions: List[Function] = []
        self.constants: List[Constant] = []

        self.initialized_globals: Dict[int, Any] = {}

        self.section_offsets: Dict[str, int] = {}

    def _find_magic(self, f: BinaryIO | BytesIO, magic: bytes = b"HLB") -> None:
        buffer_size = 1024
        offset = 0
        while True:
            chunk = f.read(buffer_size)
            if not chunk:
                raise NoMagic("Reached the end of file without finding magic bytes.")
            index = chunk.find(magic)
            if index != -1:
                f.seek(offset + index)
                dbg_print(f"Found bytecode at {tell(f)}... ", end="")
                return
            offset += buffer_size

    @classmethod
    def from_path(cls, path: str, search_magic: bool = True) -> "Bytecode":
        """
        Create a new Bytecode instance from a file path.
        """
        f = open(path, "rb")
        instance = cls().deserialise(f, search_magic=search_magic)
        f.close()
        return instance

    @classmethod
    def from_bytes(cls, data: bytes, search_magic: bool = True) -> "Bytecode":
        """
        Create a new Bytecode instance from a `bytes` object.
        """
        f = BytesIO(data)
        instance = cls().deserialise(f, search_magic=search_magic)
        f.close()
        return instance

    @classmethod
    def create_empty(cls, version: int = 4, no_extra_types: bool = False) -> "Bytecode":
        """
        Creates an empty HashLink bytecode object, ideal for adding custom functions or code to for testing. By default, contains the following types already defined:

        - t@0: Void
        - t@1: I32
        - t@2: F64
        - t@3: Bool

        Optionally, it can be created with only the Void type defined by passing `no_extra_types=True`. Note that for the bytecode to be valid and executable, there must be at least a single function, and this function must have a valid return type, meaning that the bytecode must contain **at least** 1 type.
        """
        instance = Bytecode()
        instance.magic.value = b"HLB"
        instance.version.value = version
        instance.has_debug_info = False

        void = Type()
        void.kind.value = 0  # Void (_NoDataType)
        void.definition = Void()

        if not no_extra_types:
            i32 = Type()
            i32.kind.value = 3
            i32.definition = I32()

            f64 = Type()
            f64.kind.value = 6
            f64.definition = F64()

            bool_t = Type()
            bool_t.kind.value = 7
            bool_t.definition = Bool()

        instance.types.append(void)

        if not no_extra_types:
            instance.types.append(i32)
            instance.types.append(f64)
            instance.types.append(bool_t)

        instance.set_meta()

        return instance

    def deserialise(
        self,
        f: BinaryIO | BytesIO,
        search_magic: bool = True,
        init_globals: bool = True,
    ) -> "Bytecode":
        """
        Deserialise the bytecode in-place from an open binary file handle or a BytesIO object. By default will search for the bytecode magic (b'HLB') anywhere in the file, pass `search_magic=False` to disable.
        """
        start_time = datetime.now()
        dbg_print("---- Deserialise ----")
        if search_magic:
            dbg_print("Searching for magic...")
            self._find_magic(f)
        self.track_section(f, "magic")
        self.magic.deserialise(f)
        assert self.magic.value == b"HLB", "Incorrect magic found!"
        self.track_section(f, "version")
        self.version.deserialise(f, length=1)
        dbg_print(f"with version {self.version.value}... ", end="")
        self.track_section(f, "flags")
        self.flags.deserialise(f)
        self.has_debug_info = bool(self.flags.value & 1)
        dbg_print(f"debug info: {self.has_debug_info}. ")
        self.track_section(f, "nints")
        self.nints.deserialise(f)
        self.track_section(f, "nfloats")
        self.nfloats.deserialise(f)
        self.track_section(f, "nstrings")
        self.nstrings.deserialise(f)

        if self.version.value >= 5 and self.nbytes:
            dbg_print(f"Found nbytes (version >= 5) at {tell(f)}")
            self.track_section(f, "nbytes")
            self.nbytes.deserialise(f)
        else:
            self.nbytes = None

        self.track_section(f, "ntypes")
        self.ntypes.deserialise(f)
        self.track_section(f, "nglobals")
        self.nglobals.deserialise(f)
        self.track_section(f, "nnatives")
        self.nnatives.deserialise(f)
        self.track_section(f, "nfunctions")
        self.nfunctions.deserialise(f)

        if self.version.value >= 4 and self.nconstants:
            dbg_print(f"Found nconstants (version >= 4) at {tell(f)}")
            self.track_section(f, "nconstants")
            self.nconstants.deserialise(f)
        else:
            self.nconstants = None

        self.track_section(f, "entrypoint")
        self.entrypoint.deserialise(f)
        dbg_print(f"Entrypoint: f@{self.entrypoint.value}")

        self.track_section(f, "ints")
        for i in range(self.nints.value):
            self.track_section(f, f"int {i}")
            self.ints.append(SerialisableInt().deserialise(f, length=4))

        self.track_section(f, "floats")
        for i in range(self.nfloats.value):
            self.track_section(f, f"float {i}")
            self.floats.append(SerialisableF64().deserialise(f))

        dbg_print(f"Strings section starts at {tell(f)}")
        self.track_section(f, "strings")
        self.strings.deserialise(f, self.nstrings.value)
        dbg_print(f"Strings section ends at {tell(f)}")
        assert self.nstrings.value == len(self.strings.value), "nstrings and len of strings don't match!"

        if self.version.value >= 5 and self.bytes and self.nbytes:
            dbg_print("Deserialising bytes... >=5")
            self.track_section(f, "bytes")
            self.bytes.deserialise(f, self.nbytes.value)
        else:
            self.bytes = None

        if self.has_debug_info and self.ndebugfiles and self.debugfiles:
            dbg_print(f"Deserialising debug files... (at {tell(f)})")
            self.track_section(f, "ndebugfiles")
            self.ndebugfiles.deserialise(f)
            dbg_print(f"Number of debug files: {self.ndebugfiles.value}")
            self.track_section(f, "debugfiles")
            self.debugfiles.deserialise(f, self.ndebugfiles.value)
        else:
            self.ndebugfiles = None
            self.debugfiles = None

        dbg_print(f"Starting main blobs at {tell(f)}")
        dbg_print(f"Types starting at {tell(f)}")
        self.track_section(f, "types")
        for i in range(self.ntypes.value):
            self.track_section(f, f"type {i}")
            self.types.append(Type().deserialise(f))
        dbg_print(f"Globals starting at {tell(f)}")
        self.track_section(f, "globals")
        for i in range(self.nglobals.value):
            self.track_section(f, f"global {i}")
            self.global_types.append(tIndex().deserialise(f))
        dbg_print(f"Natives starting at {tell(f)}")
        self.track_section(f, "natives")
        for i in range(self.nnatives.value):
            self.track_section(f, f"native {i}")
            self.natives.append(Native().deserialise(f))
        dbg_print(f"Functions starting at {tell(f)}")
        self.track_section(f, "functions")
        if not USE_TQDM or self.nfunctions.value < 1000:
            for i in range(self.nfunctions.value):
                self.track_section(f, f"function {i}")
                self.functions.append(Function().deserialise(f, self.has_debug_info, self.version.value))
        else:
            for i in tqdm(range(self.nfunctions.value)):
                self.track_section(f, f"function {i}")
                self.functions.append(Function().deserialise(f, self.has_debug_info, self.version.value))
        if self.nconstants is not None:
            dbg_print(f"Constants starting at {tell(f)}")
            self.track_section(f, "constants")
            for i in range(self.nconstants.value):
                self.track_section(f, f"constant {i}")
                self.constants.append(Constant().deserialise(f))
        dbg_print(f"Bytecode end at {tell(f)}.")
        self.deserialised = True
        if init_globals:
            dbg_print("Initializing globals...")
            self.init_globals()
        dbg_print(f"{(datetime.now() - start_time).total_seconds()}s elapsed.")
        return self

    def init_globals(self) -> None:
        """
        Internal method to initialize global instances of objects.
        """
        final: Dict[int, Any] = {}
        for const in self.constants:
            res: Dict[str, Any] = {}
            obj = const._global.resolve(self).definition
            if not isinstance(obj, Obj):
                dbg_print("WARNING: Skipping non-Obj constant.")
                continue
            obj_fields = obj.resolve_fields(self)
            for i, field in enumerate(const.fields):
                # Field has:
                # - name: strRef
                # - type: tIndex
                # we need to use the type to know how to resolve the const ref to the actual value
                typ = obj_fields[i].type.resolve(self).definition
                name = obj_fields[i].name.resolve(self)
                if isinstance(typ, I32) or isinstance(typ, U8) or isinstance(typ, U16) or isinstance(typ, I64):
                    res[name] = self.ints[field.value].value
                elif isinstance(typ, F32) or isinstance(typ, F64):
                    res[name] = self.floats[field.value].value
                elif isinstance(typ, Bytes):
                    res[name] = self.strings.value[field.value]
                else:
                    res[name] = field.value
            final[const._global.value] = res
        assert len(final) == len(self.constants), (
            "Not all constants were resolved! This is often due to bad DebugInfo blocks causing buffer overrun, try passing -N to troubleshoot."
        )
        self.initialized_globals = final

    def fn(self, findex: int, native: bool = True) -> Function | Native:
        """
        Shorthand to to get a Function or a Native by its fIndex.
        """
        for f in self.functions:
            if f.findex.value == findex:
                return f
        if native:
            for n in self.natives:
                if n.findex.value == findex:
                    return n
        raise ValueError(f"Function {findex} not found!")

    def t(self, tindex: int) -> Type:
        """
        Shorthand to get a Type by its index. Equivalent to code.types[tindex]
        """
        return self.types[tindex]

    def g(self, gindex: int) -> Type:
        """
        Shorthand to get a global's type by gIndex.
        """
        for g in self.global_types:
            if g.value == gindex:
                return g.resolve(self)
        raise ValueError(f"Global {gindex} not found!")

    def const_str(self, gindex: int) -> str:
        """
        Gets the value of an initialized global constant `String`.
        """
        # TODO: is this overcomplicated?
        if gindex not in self.initialized_globals:
            if gindex < 0 or gindex >= len(self.global_types):
                raise ValueError(f"Global {gindex} not found!")
            raise ValueError(f"Global {gindex} does not have a constant value!")
        obj = self.global_types[gindex].resolve(self).definition
        if not isinstance(obj, Obj):
            raise TypeError(f"Global {gindex} is not an object!")
        if not obj.name.resolve(self) == "String":
            raise TypeError(f"Global {gindex} is not a string!")
        obj_fields = obj.resolve_fields(self)
        if len(obj_fields) != 2:
            raise ValueError(f"Global {gindex} seems malformed!")
        res = self.initialized_globals[gindex][obj_fields[0].name.resolve(self)]
        if not isinstance(res, str):
            raise TypeError("This should never happen!")
        return res

    def serialise(self, auto_set_meta: bool = True) -> bytes:
        """
        Serialise the bytecode to a `bytes` object.
        """
        start_time = datetime.now()
        dbg_print("---- Serialise ----")
        if auto_set_meta:
            dbg_print("Setting meta...")
            self.set_meta()
        res = b"".join(
            [
                self.magic.serialise(),
                self.version.serialise(),
                self.flags.serialise(),
                self.nints.serialise(),
                self.nfloats.serialise(),
                self.nstrings.serialise(),
            ]
        )
        dbg_print(f"VarInt block 1 at {hex(len(res))}")
        if self.version.value >= 5 and self.nbytes:
            res += self.nbytes.serialise()
        res += b"".join(
            [
                self.ntypes.serialise(),
                self.nglobals.serialise(),
                self.nnatives.serialise(),
                self.nfunctions.serialise(),
            ]
        )
        dbg_print(f"VarInt block 2 at {hex(len(res))}")
        if self.version.value >= 4 and self.nconstants:
            res += self.nconstants.serialise()
        res += self.entrypoint.serialise()
        res += b"".join([i.serialise() for i in self.ints])
        res += b"".join([f.serialise() for f in self.floats])
        res += self.strings.serialise()
        if self.version.value >= 5 and self.bytes:
            res += self.bytes.serialise()
        if self.has_debug_info and self.ndebugfiles and self.debugfiles:
            res += b"".join([self.ndebugfiles.serialise(), self.debugfiles.serialise()])
        res += b"".join(
            [
                b"".join([typ.serialise() for typ in self.types]),
                b"".join([typ.serialise() for typ in self.global_types]),
                b"".join([native.serialise() for native in self.natives]),
            ]
        )
        if USE_TQDM:
            res += b"".join([func.serialise() for func in tqdm(self.functions)])
        else:
            res += b"".join([func.serialise() for func in self.functions])
        res += b"".join([constant.serialise() for constant in self.constants])
        dbg_print(f"Final size: {hex(len(res))}")
        dbg_print(f"{(datetime.now() - start_time).total_seconds()}s elapsed.")
        return res

    def set_meta(self) -> None:
        """
        Sets bytecode metadata automatically.
        """
        self.flags.value = 1 if self.has_debug_info else 0
        self.nints.value = len(self.ints)
        self.nfloats.value = len(self.floats)
        self.nstrings.value = len(self.strings.value)
        if self.version.value >= 5 and self.bytes and self.nbytes:
            self.nbytes.value = len(self.bytes.value)
        self.ntypes.value = len(self.types)
        self.nglobals.value = len(self.global_types)
        self.nnatives.value = len(self.natives)
        self.nfunctions.value = len(self.functions)
        if self.version.value >= 4 and self.nconstants:
            self.nconstants.value = len(self.constants)
        if self.has_debug_info and self.ndebugfiles and self.debugfiles:
            self.ndebugfiles.value = len(self.debugfiles.value)

    def repair(self) -> None:
        """
        Alternate notation for code.set_meta() for the purpose of clarity.
        """
        self.set_meta()

    def get_test_main(self) -> Function:
        for f in self.functions:
            if full_func_name(self, f).endswith("main"):
                return f
        raise ValueError("No main function found!")

    def is_ok(self) -> bool:
        """
        Runs a set of basic sanity checks to make sure the bytecode is correct-ish.
        """

        def fail(msg: str) -> None:
            print(f"--- FAILED CHECK ---\n{msg}")

        if len(self.ints) != self.nints.value:
            fail("ints != nints")
            return False

        if len(self.floats) != self.nfloats.value:
            fail("floats != nfloats")
            return False

        if len(self.strings.value) != self.nstrings.value:
            fail("strings != nstrings")
            print(len(self.strings.value), self.nstrings.value)
            return False

        if self.version.value >= 5:
            if self.nbytes is None or self.bytes is None:
                fail("nbytes or bytes is None and version >= 5")
                return False
            if len(self.bytes.value) != self.nbytes.value:
                fail("bytes != nbytes")
                return False

        if len(self.types) != self.ntypes.value:
            fail("types != ntypes")
            return False

        if len(self.global_types) != self.nglobals.value:
            fail("globals != nglobals")
            return False

        if len(self.natives) != self.nnatives.value:
            fail("natives != nnatives")
            return False

        if len(self.functions) != self.nfunctions.value:
            fail("functions != nfunctions")
            return False

        if self.version.value >= 4:
            if self.nconstants is None:
                fail("nconstants is None and version >= 4")
                return False
            if len(self.constants) != self.nconstants.value:
                fail("constants != nconstants")
                return False

        if self.has_debug_info:
            if self.ndebugfiles is None or self.debugfiles is None:
                fail("ndebugfiles or debugfiles is None and has_debug_info")
                return False
            if len(self.debugfiles.value) != self.ndebugfiles.value:
                fail("debugfiles != ndebugfiles")
                return False

        return True

    def track_section(self, f: BinaryIO | BytesIO, section_name: str) -> None:
        """
        Internal helper function to denote the location of a data section at a given offset.
        """
        self.section_offsets[section_name] = f.tell()

    def section_at(self, offset: int) -> Optional[str]:
        """
        Returns the name of the bytecode data section at the offset.
        """
        # returns the name of the section at the offset:
        # if the offset is after a section start and before the next section start, it's still in the first section
        for section_name, section_offset in list(reversed(self.section_offsets.items())):
            if offset >= section_offset:
                return section_name
        return None

    def add_string(self, string: str) -> strRef:
        """
        Adds a string to the bytecode's string block and returns a reference to it.
        """
        return strRef(self.strings.find_or_add(string))

    def add_i32(self, value: int) -> intRef:
        """
        Adds an integer to the bytecode's integer block and returns a reference to it.
        """
        val = SerialisableInt()
        val.value = value
        self.ints.append(val)
        return intRef(len(self.ints) - 1)

    def next_free_findex(self) -> fIndex:
        """
        Find the next available fIndex that is not already used by any function or native.
        """
        used_indexes = set()
        for function in self.functions:
            used_indexes.add(function.findex.value)
        for native in self.natives:
            used_indexes.add(native.findex.value)

        index = 0
        while index in used_indexes:
            index += 1

        return fIndex(index)

    def find_prim_type(self, kind: Type.Kind) -> tIndex:
        """
        Finds the index of a primitive type in the bytecode.
        """
        assert kind.value in [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            12,
            13,
            16,
        ], f"This method can only find primitive types! Got: {kind}"
        for i, typ in enumerate(self.types):
            if typ.kind.value == kind.value:
                return tIndex(i)
        raise ValueError(f"Primitive type {kind} not found!")

    def add_type(self, typ: Type) -> tIndex:
        """
        Adds a type and returns its tIndex.
        """
        self.types.append(typ)
        return tIndex(len(self.types) - 1)


def get_field_for(code: Bytecode, idx: int) -> Optional[Field]:
    """
    Gets the field for a standalone function index.
    """
    for type in code.types:
        if type.kind.value == Type.TYPEDEFS.index(Obj):
            if not isinstance(type.definition, Obj):
                raise TypeError(f"Expected Obj, got {type.definition}")
            definition: Obj = type.definition
            fields = definition.resolve_fields(code)
            for binding in definition.bindings:  # binding binds a field to a function
                if binding.findex.value == idx:
                    return fields[binding.field.value]
    return None


def get_proto_for(code: Bytecode, idx: int) -> Optional[Proto]:
    """
    Gets the proto for a standalone function index.
    """
    for type in code.types:
        if type.kind.value == Type.TYPEDEFS.index(Obj):
            if not isinstance(type.definition, Obj):
                raise TypeError(f"Expected Obj, got {type.definition}")
            definition: Obj = type.definition
            for proto in definition.protos:
                if proto.findex.value == idx:
                    return proto
    return None


def full_func_name(code: Bytecode, func: Function | Native) -> str:
    """
    Generates a human-readable name for a function or native.
    """
    proto = get_proto_for(code, func.findex.value)
    if proto:
        name = proto.name.resolve(code)
        for type in code.types:
            if type.kind.value == Type.TYPEDEFS.index(Obj):
                if not isinstance(type.definition, Obj):
                    continue
                obj_def: Obj = type.definition
                for fun in obj_def.protos:
                    if fun.findex.value == func.findex.value:
                        return f"{obj_def.name.resolve(code)}.{name}"
    else:
        name = "<none>"
        field = get_field_for(code, func.findex.value)
        if field:
            name = field.name.resolve(code)
            for type in code.types:
                if type.kind.value == Type.TYPEDEFS.index(Obj):
                    if not isinstance(type.definition, Obj):
                        continue
                    _obj_def: Obj = type.definition
                    fields = _obj_def.resolve_fields(code)
                    for binding in _obj_def.bindings:
                        if binding.findex.value == func.findex.value:
                            return f"{_obj_def.name.resolve(code)}.{name}"
    return name


def partial_func_name(code: Bytecode, func: Function | Native) -> str:
    """
    Generates a human-readable name for a function or native. Does not qualify the name with the object it belongs to.
    """
    proto = get_proto_for(code, func.findex.value)
    if proto:
        return proto.name.resolve(code)
    else:
        field = get_field_for(code, func.findex.value)
        if field:
            return field.name.resolve(code)
    return "<none>"


__all__ = [
    "full_func_name",
    "get_field_for",
    "get_proto_for",
    "Abstract",
    "Array",
    "Binding",
    "Bool",
    "Bytecode",
    "Bytes",
    "BytesBlock",
    "Constant",
    "DebugInfo",
    "Dyn",
    "DynObj",
    "Enum",
    "EnumConstruct",
    "F32",
    "F64",
    "Field",
    "Fun",
    "Function",
    "I32",
    "I64",
    "InlineBool",
    "Method",
    "Native",
    "Null",
    "Obj",
    "Opcode",
    "Packed",
    "Proto",
    "RawData",
    "Ref",
    "Reg",
    "Regs",
    "ResolvableVarInt",
    "Serialisable",
    "SerialisableF64",
    "SerialisableInt",
    "StringsBlock",
    "Struct",
    "Type",
    "TypeDef",
    "TypeType",
    "U16",
    "U8",
    "VarInt",
    "VarInts",
    "Virtual",
    "Void",
    "bytesRef",
    "fIndex",
    "fieldRef",
    "fileRef",
    "floatRef",
    "gIndex",
    "intRef",
    "strRef",
    "tIndex",
]



"""Pure-Python HashLink bytecode Swiss Army knife.

> [!WARNING]
> This project is under active development. Breaking changes may be made to APIs with zero notice.

## Features

- Pure Python with zero dependencies, integrates nicely in a lot of places (IDAPython compatible!)
- Deserialisation, disassembly, and (currently limited) decompilation of HashLink bytecode
- Reserialisation and first-class support for patching bytecode assembly
- A bytecode assembler for creating HashLink bytecode from scratch
- A scriptable interface for easy integration into other tools
- A little CLI similar to [hlbc](https://github.com/Gui-Yom/hlbc)

## Installation

```bash
pip install crashlink
```

Optionally, install `[extras]` for progress bars when parsing large files and faster bytecode save/load in-memory:

```bash
pip install crashlink[extras]
```

Or, for bleeding-edge features, see the [Development](#development) section.

You also need to have Graphviz installed to generate control flow graphs. On most *nix systems, on Windows (with Chocolatey or Scoop), and on MacOS (with Homebrew), you can install it with your package manager under `graphviz`.

- Windows: `choco install graphviz`
- MacOS: `brew install graphviz`
- Debian: `sudo apt install graphviz`
- Arch: `sudo pacman -S graphviz`
- Fedora: `sudo dnf install graphviz`

## Usage

Either:

```txt
$ crashlink path/to/file.hl # or python -m crashlink
crashlink> funcs
f@22 static Clazz.main () -> Void (from Clazz.hx)
f@23 Clazz.method (Clazz) -> I32 (from Clazz.hx)
crashlink> fn 22
f@22 static Clazz.main () -> Void (from Clazz.hx)
Reg types:
  0. Void

Ops:
  0. Ret             {'ret': 0}                                       return
```

Or:

```py
from crashlink import *
code = Bytecode.from_path("path/to/file.hl")
if code.fn(22): # 22 and 240 are typical entry points for the compiler to generate
  print(disasm.func(code.fn(22)))
elif code.fn(240):
  print(disasm.func(code.fn(240)))
# > f@22 static $Clazz.main () -> Void (from Clazz.hx)
# > Reg types:
# >   0. Void
# >
# > Ops:
# >   0. Ret             {'ret': 0}                                       return
```

Read the [API documentation](https://n3rdl0rd.github.io/crashlink/crashlink) for more information.

## Development

> [!NOTE]
> This project is configured for the [just](https://just.systems/) command runner and [uv](https://docs.astral.sh/uv/). If you don't have them installed, you can still run the commands in the `justfile` manually, but I don't recommend it.

For development purposes, you can clone the repo, install development dependencies, and run the tests:

```bash
git clone https://github.com/N3rdL0rd/crashlink
cd crashlink
# optionally, create and activate a venv here.
just install # or uv pip install -e .[dev]
just test # or pytest
```

Before committing, please run `just dev` to format the code, run tests, and generate documentation in `docs/`. If you're adding new features to the core serialisation/deserialisation code (`core.py`), please also add a test case in `tests/haxe/` for the new language feature you're adding. If you're adding a feature to the decompiler or disassembler, please add a normal test case (in Python) in `tests/` that tests the new feature.

Pull requests are always welcome! For major changes, please open an issue first to discuss what you would like to change.

You can use the following pre-defined commands with `just`:

- `just dev`: Run tests, format code, and generate documentation.
- `just build`: Build the package.
- `just install`: Install development dependencies and the package in editable mode.
- `just build-tests`: Build test samples.
- `just test`: Run tests.
- `just format`: Format code.
- `just docs`: Generate documentation.
- `just check`: Run static analysis/typechecking.
- `just clean`: Clean up build artifacts.
- `just profile`: Run the test suite with cProfile and then open the results in a browser.
- `just serve-docs`: Serve the documentation locally.

> [!NOTE]
> A lot of very WIP development is done on `dev`. If you want to see the latest and greatest, checkout that branch. If you want something semi-stable, checkout `main` instead.

## Roadmap

- [x] Bytecode parsing
- [x] Opcode disassembly
  - [x] Local resolution and naming
- [ ] IR lifter (layer 0)
  - [x] If statements
  - [ ] Loops
  - [x] Switch opcode statements
  - [ ] Function calls
    - [ ] CallClosure
  - [ ] Closures, lambdas
- [ ] IR optimization layers
  - [ ] SSA locals
  - [ ] Trace optimization
  - [ ] Nested if/else/if/else -> switch
- [ ] Pseudocode targets
  - [ ] Haxe (main target)
  - [ ] Any others? We'll see...
- [ ] Partial recompilation (against stubs of other functions)
- [ ] GUI? (customtkinter or dearpygui)
  - [ ] Graphical disassembler
  - [ ] Embedded CFG viewer through some Graphviz bindings
  - [ ] Decompiler and patching interface
  - [ ] IR layer viewer

## Portability

crashlink is written in pure typed Python with a minimum version of 3.10 (for the `|` operator and `match` statement). It should run on any modern platform, and has been tested heavily on Windows, Linux, and MacOS. As well as this, it is portable to many interpreters:

- CPython 3.10+ is the main target - deserialising Dead Cells v35 takes 103.16s and 4.6gb of RAM
- PyPy just works - the same benchmark takes 26s and 2.9gb of RAM
  - tqdm displays broken progress bars, but this is a PyPy issue and it still works.
- IronPython and Jython are not supported due to their earlier Python version targets.
- RustPython would work, but it doesn't support `match` statements.

## Credits

- Thank you to [Gui-Yom](https://github.com/Gui-Yom) for writing hlbc and for maintaining documentation on the HashLink bytecode format, as well as for providing tests and helping me during development.
- Thank you to [Haxe Foundation](https://haxe.org/) for creating the HashLink VM and the Haxe programming language.
- Thank you to the Dead Cells community on Discord for providing me with the motivation to start this project.
- And a big thank you to you, dear user, for being at least partially interested in this project.

❤ N3rdL0rd

"""

from __future__ import annotations

from .asm import *
from .core import *
from .errors import *
from .globals import *
from .opcodes import *
from . import decomp as decomp
from . import disasm as disasm
from . import pseudo as pseudo
from . import interp as interp



from glob import glob

import pytest

from crashlink import *

test_files = glob("tests/haxe/*.hl")


@pytest.mark.parametrize("path", test_files)
def test_deser_basic(path: str):
    with open(path, "rb") as f:
        code = Bytecode().deserialise(f)
        assert code.is_ok()


@pytest.mark.parametrize("path", test_files)
def test_reser_basic(path: str):
    with open(path, "rb") as f:
        code = Bytecode().deserialise(f)
        assert code.is_ok(), "Failed during deser"
        f.seek(0)
        if not f.read() == code.serialise():
            # find 1st non-matching byte
            ser = code.serialise()
            f.seek(0)
            c = 0
            msg = ""
            while True:
                a = f.read(1)
                b = ser[c : c + 1]
                t = f.tell() - 1
                if a != b:
                    msg = f"First mismatch at {hex(t)}: {a!r} != {b!r} (in section '{code.section_at(t)}')"
                    print(msg)
                    break
                c += 1
            assert False, "Failed matching reser: " + msg


def test_create_empty():
    code = Bytecode.create_empty()
    assert code.is_ok(), "Bad code!"
    assert Bytecode.from_bytes(code.serialise()).serialise() == code.serialise()



"""
Functions to build the crashtest results site.
"""

import os

from staticjinja import Site  # type: ignore

from .models import load_runs


def build() -> None:
    """
    Build the site.
    """
    template_path = os.path.join(os.path.dirname(__file__), "templates")
    runs = load_runs(os.path.join(os.path.dirname(__file__), "runs"))
    contexts = [
        (
            "index.html",
            {
                "runs": runs,
            },
        )
    ]
    site = Site.make_site(searchpath=template_path, outpath="docs/crashtest_out", contexts=contexts)
    for run in runs:
        context = {"context": run.context, "git": run.git, "cases": run.cases}

        template = site.get_template("_result.html")
        output_path = f"docs/crashtest_out/run/{run.id}/index.html"
        site.render_template(template, context=context, filepath=output_path)
    site.render()



"""
Functions to run tests, collect results, and produce run reports.
"""

import datetime
import os
import re
import subprocess
import traceback
from typing import List, Tuple
from markupsafe import escape

from crashlink import Bytecode, decomp, globals
from crashlink.pseudo import pseudo

from .models import GitInfo, Run, TestCase, TestContext, TestFile, save_run


def get_repo_info() -> GitInfo:
    """
    Get the git branch and commit hash, if available.
    """
    try:
        original_dir = os.getcwd()
        script_dir = os.path.dirname(os.path.abspath(__file__))
        os.chdir(script_dir)

        try:
            branch = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).strip().decode("utf-8")
            commit = subprocess.check_output(["git", "rev-parse", "HEAD"]).strip().decode("utf-8")
            dirty = subprocess.check_output(["git", "status", "--porcelain"]).strip().decode("utf-8") != ""
            return GitInfo(
                is_release=False,
                branch=branch,
                commit=commit[:8],
                dirty=dirty,
                github=f"https://github.com/N3rdL0rd/crashlink/commit/{commit}",
            )
        finally:
            os.chdir(original_dir)
    except (subprocess.CalledProcessError, FileNotFoundError):
        return GitInfo(is_release=True, dirty=False)


def file_to_name(file: str) -> str:
    return " ".join(
        re.sub(
            "([A-Z][a-z]+)",
            r" \1",
            re.sub("([A-Z]+)", r" \1", file.replace(".hx", "").replace("_", " ")),
        ).split()
    ).title()


def run_case(case: str, id: int) -> TestCase:
    """
    Runs a single test case, handling IR and pseudocode generation separately.
    """
    # Attempt to read the original content
    try:
        original_content = open(
            os.path.join(os.path.dirname(__file__), "..", "tests", "haxe", case),
            "r",
        ).read()
    except Exception as e:
        tb_last = traceback.format_exc().splitlines()[-1]
        return TestCase(
            original=TestFile(
                name=case,
                content=escape("Failed to read original file."),
            ),
            decompiled=TestFile(
                name=f"{case.replace('.hx', '')} (Decompiled)",
                content=escape("Failed to produce pseudocode."),
            ),
            ir=TestFile(name=f"{case.replace('.hx', '')} (IR)", content=escape("Failed to produce IR.")),
            failed=True,
            test_name=file_to_name(case),
            test_id=id,
            error=escape(f"Failed to read original file: {str(e)}\n{tb_last}"),
        )

    # Load bytecode and create IRFunction
    code = None
    irf = None
    ir_content = "Failed to produce IR."
    pseudo_content = "Failed to produce pseudocode."
    ir_error = None
    pseudo_error = None

    # First try to load the bytecode
    try:
        code = Bytecode.from_path(
            os.path.join(
                os.path.dirname(__file__),
                "..",
                "tests",
                "haxe",
                case.replace(".hx", ".hl"),
            )
        )
        irf = decomp.IRFunction(code, code.get_test_main())
    except Exception as e:
        tb_last = traceback.format_exc().splitlines()[-1]
        ir_error = f"Failed to load bytecode: {str(e)}\n{tb_last}"
        pseudo_error = f"Failed to load bytecode: {str(e)}\n{tb_last}"
        return TestCase(
            original=TestFile(
                name=case,
                content=escape(original_content),
            ),
            decompiled=TestFile(
                name=f"{case.replace('.hx', '')} (Decompiled)",
                content=escape(pseudo_content),
            ),
            ir=TestFile(name=f"{case.replace('.hx', '')} (IR)", content=escape(ir_content)),
            failed=True,
            test_name=file_to_name(case),
            test_id=id,
            error=escape(ir_error),
        )

    # Try to generate IR
    try:
        if irf:
            ir_content = str(irf.block)
    except Exception as e:
        tb_last = traceback.format_exc().splitlines()[-1]
        ir_error = f"Failed to generate IR: {str(e)}\n{tb_last}"

    # Try to generate pseudocode
    try:
        if irf:
            pseudo_content = pseudo(irf)
    except Exception as e:
        tb_last = traceback.format_exc().splitlines()[-1]
        pseudo_error = f"Failed to generate pseudocode: {str(e)}\n{tb_last}"

    # Determine if the test failed based on errors
    failed = bool(ir_error and pseudo_error)

    # Create the error message with all available information
    error = None
    if ir_error or pseudo_error:
        error_parts = []
        if ir_error:
            error_parts.append(ir_error)
        if pseudo_error:
            error_parts.append(pseudo_error)
        error = escape("\n".join(error_parts))

    return TestCase(
        original=TestFile(
            name=case,
            content=escape(original_content),
        ),
        decompiled=TestFile(
            name=f"{case.replace('.hx', '')} (Decompiled)",
            content=escape(pseudo_content),
        ),
        ir=TestFile(name=f"{case.replace('.hx', '')} (IR)", content=escape(ir_content)),
        failed=failed,
        test_name=file_to_name(case),
        test_id=id,
        error=error,
    )


def gen_id() -> str:
    """
    Generate a unique ID for a test run.
    """
    return datetime.datetime.now().strftime("%Y%m%d%H%M%S")


def gen_status(results: List[TestCase]) -> Tuple[str, str]:
    """
    Generate a status message and color based on test results.
    Returns a tuple of (status_message, color_hex).

    Colors:
    - Green (#22C55E): All tests passed
    - Yellow (#EAB308): < 10% failures
    - Orange (#F97316): 10-20% failures
    - Red-Orange (#EF4444): 20-50% failures
    - Red (#DC2626): > 50% failures
    - Dark Red (#991B1B): All tests failed
    """
    if not results:
        return "No Tests Run", "#6B7280"  # Gray for no tests

    total = len(results)
    failed = sum(1 for case in results if case.failed)
    failure_rate = (failed / total) * 100

    if failed == 0:
        return "All tests passed", "#22C55E"
    elif failed == total:
        return "All tests failed", "#991B1B"
    else:
        if failure_rate < 10:
            return f"Partial failure ({failure_rate:.1f}%)", "#EAB308"
        elif failure_rate < 20:
            return f"Partial failure ({failure_rate:.1f}%)", "#F97316"
        elif failure_rate < 50:
            return f"Major failures ({failure_rate:.1f}%)", "#EF4444"
        else:
            return f"Critical failures ({failure_rate:.1f}%)", "#DC2626"


def run() -> None:
    """
    Run all tests.
    """
    print("Getting repo info...")
    git = get_repo_info()
    if git.is_release:
        print(
            "Cannot run tests from a release build (eg. installed fro PyPI). Please clone the repo and run from there."
        )
        return  # TODO: add support for autodownloading and building test samples

    print("Finding test cases...")
    files = os.listdir(os.path.join(os.path.dirname(__file__), "..", "tests", "haxe"))
    cases = [f for f in files if f.endswith(".hx")]
    for case in cases:
        if case.replace(".hx", ".hl") not in files:
            print(f"Warning: no compiled bytecode found for {case}. Skipping.")
            cases.remove(case)

    print("Running tests...")
    results = []
    for i, case in enumerate(cases):
        print(f"Running {case}...")
        result = run_case(case, i)
        results.append(result)

    print("Generating run...")
    status, status_color = gen_status(results)
    r = Run(
        git=git,
        context=TestContext(version=globals.VERSION),
        cases=results,
        id=gen_id(),
        timestamp=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        status=status,
        status_color=status_color,
    )
    os.makedirs(os.path.join(os.path.dirname(__file__), "runs"), exist_ok=True)
    save_run(r, os.path.join(os.path.dirname(__file__), "runs", f"{gen_id()}.json"))



import os

import pytest

from crashlink import *

POSSIBLE_LOCATIONS = [
    "C:\\Program Files (x86)\\Steam\\steamapps\\common\\Dead Cells\\deadcells.exe",
    "C:\\Program Files\\Steam\\steamapps\\common\\Dead Cells\\deadcells.exe",
    "D:\\SteamLibrary\\steamapps\\common\\Dead Cells\\deadcells.exe",
    "E:\\SteamLibrary\\steamapps\\common\\Dead Cells\\deadcells.exe",
    "F:\\SteamLibrary\\steamapps\\common\\Dead Cells\\deadcells.exe",
    "~/.steam/steam/steamapps/common/Dead\\ Cells/hlboot.dat",
]


@pytest.mark.skip(reason="Not open-source. If you own the game, feel free to run this test.")
def test_deser_deadcells():
    loc = next((loc for loc in POSSIBLE_LOCATIONS if os.path.exists(loc)), None)
    if loc is None:
        print("Dead Cells not found. Skipping test.")
        assert True
        return
    code = Bytecode.from_path(loc)
    assert code.is_ok()



# type: ignore
# ^ Currently flawed and not ready for typing. Ignore it for now, and fix it later.

"""
Port of HashLink's std C bindings to pure Python. Entirely functional due to the nature of the library and C bindings.
"""

from dataclasses import dataclass
from typing import Any, Callable, List

from ..core import Bytecode, Obj


@dataclass
class StdBinding:
    """
    Represents a binding between a Haxe/HashLink standard library function name and its Python implementation.
    """

    name: str
    lib: str
    func: Callable[..., Any]


BINDINGS: List[StdBinding] = []


def bind(name: str, lib: str) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Decorator to bind a Python function to a Haxe/HashLink standard library function name.
    """

    def inner(func: Callable[..., Any]) -> Callable[..., Any]:
        BINDINGS.append(StdBinding(name, lib, func))
        return func

    return inner


@bind("hballoc", "std")
def hballoc(code: Bytecode, vm: "VM") -> "VMBytes":
    """
    Allocate a byte array.
    """
    from .vm import VMBytes  # deferred import, kill me

    return VMBytes(b"")


@bind("type_get_global", "std")
def type_get_global(code: Bytecode, vm: "VM", typ: "VMType") -> "VMPrim":
    """
    Get a global instance of an object type, if one is present.
    Does nothing when passed an Obj, only works with Structs and Enums.
    """
    match typ.value.kind.value:
        case 11:  # Obj - explicitly ignored
            pass
        case 21 | 18:  # Struct | Enum - both have the same attribute "_global" so they should be interchangable (?)
            if typ.value.definition._global and vm.globals[typ.value.definition._global.value]:
                return vm.globals[typ.value.definition._global.value]
            raise NameError(f"Global instance of {typ} not found.")
    return None


@bind("type_set_global", "std")
def type_set_global(code: Bytecode, vm: "VM", typ: "VMType", instance: "VMValue") -> "VMBool":
    """
    Set a global instance of an object type to the given instance or Obj (Dyn) - types are unchecked so be careful, YOLO!
    """
    from .vm import VMBool

    match typ.value.kind.value:
        case 11:  # Obj - explicitly ignored
            pass
        case 21 | 18:  # Struct | Enum
            if typ.value.definition._global:
                vm.globals[typ.value.definition._global.value] = instance
                return VMBool(True)
            raise NameError(f"Global instance of {typ} not found.")
    return VMBool(False)


@bind("alloc_obj", "std")
def alloc_obj(code: Bytecode, vm: "VM", typ: "VMType") -> "VMValue":
    """
    Allocate an empty Obj of the given type.
    """
    from .vm import VMObj

    obj = typ.value.definition
    assert isinstance(obj, Obj), "Not an Obj."
    return VMObj.create_empty(obj, code)


@bind("sys_print", "std")
def sys_print(code: Bytecode, vm: "VM", value: "VMValue") -> "VMVoid":
    """
    Print a value to the console.
    """
    from .vm import VMBytes

    assert isinstance(value, VMBytes), "Can only print bytes."
    print(value.value.decode("utf-8"), end="")
    return None



"""
Prettier HashLink bytecode notation.
"""

from __future__ import annotations

import re
from abc import ABC
from dataclasses import dataclass, field
from typing import Any, Dict, List

from .core import (
    F32,
    F64,
    I32,
    I64,
    U8,
    U16,
    Array,
    Bool,
    Bytecode,
    Bytes,
    Dyn,
    Fun,
    Function,
    Native,
    Opcode,
    Reg,
    ResolvableVarInt,
    Type,
    TypeType,
    Void,
    bytesRef,
    fIndex,
    gIndex,
    intRef,
    strRef,
    tIndex,
)
from .opcodes import opcodes


@dataclass
class AsmValue(ABC):
    value: Any


class AsmValueStr(AsmValue):
    value: str


@dataclass
class AsmSection(AsmValue):
    name: str = ""
    value: "List[AsmValueStr|AsmSection]" = field(default_factory=list)

    def get(self, subsection_name: str) -> "AsmSection":
        for val in self.value:
            if isinstance(val, AsmSection) and val.name == subsection_name:
                return val
        raise KeyError(f"No subsection '{subsection_name}' found!")


class AsmFile:
    def __init__(self, content: str) -> None:
        self.content = content
        self.raw_sections: Dict[str, AsmSection] = {}
        self.strings: List[str] = []
        self._parse()

    @classmethod
    def from_path(cls, path: str) -> "AsmFile":
        with open(path, "r", encoding="utf-8") as file:
            content = file.read()
        return cls(content)

    def _parse(self) -> None:
        self.content = self.content.replace("    ", "\t")  # for consistency
        lines = self.content.splitlines()
        section_stack: List[AsmSection] = []
        for line in lines:
            if not line.strip() or line.strip().startswith("#"):
                continue
            indent_level = len(line) - len(line.lstrip("\t"))
            # pop extra sections if we decreased the indent level
            while len(section_stack) > indent_level:
                section_stack.pop()
            stripped = line.lstrip("\t")
            if stripped.startswith("."):
                tokens = stripped.split()
                section_name = tokens[0][1:]
                new_section = AsmSection(value=[])
                new_section.name = section_name
                if len(tokens) > 1:
                    for token in tokens[1:]:
                        new_section.value.append(AsmValueStr(token))
                if section_stack:
                    section_stack[-1].value.append(new_section)
                else:
                    self.raw_sections[section_name] = new_section
                section_stack.append(new_section)
            else:
                if not section_stack:
                    raise SyntaxError("Encountered a value outside any section!")
                section_stack[-1].value.append(AsmValueStr(stripped))

    def _add_types(self, code: Bytecode, section: AsmSection) -> None:
        name_to_def = {
            "Void": Void,
            "U8": U8,
            "U16": U16,
            "I32": I32,
            "I64": I64,
            "F32": F32,
            "F64": F64,
            "Bool": Bool,
            "Bytes": Bytes,
            "Dyn": Dyn,
            "Array": Array,
            "Type": TypeType,
        }
        def_to_kind = {
            Void: 0,
            U8: 1,
            U16: 2,
            I32: 3,
            I64: 4,
            F32: 5,
            F64: 6,
            Bool: 7,
            Bytes: 8,
            Dyn: 9,
            Array: 12,
            TypeType: 13,
        }
        for val in section.value:
            print(val.value)
            if not isinstance(val, AsmValueStr):
                continue
            parts = val.value.split()
            if parts[0] in name_to_def:
                typedef = name_to_def[parts[0]]
                m_def = typedef()
                typ = Type()
                typ.kind.value = def_to_kind[typedef]
                typ.definition = m_def
                code.types.append(typ)
            elif parts[0] == "Fun":
                print("Adding Fun...")
                fun = Fun()
                tokens = re.findall(r"\([^)]*\)|\S+", val.value)
                _, args, _, ret = tokens
                r = self._parse_ref(ret)
                assert isinstance(r, tIndex), "Expected a type reference for return!"
                fun.ret = r
                args_s = args.strip("()").split(",")
                if len(args_s) == 1 and not args_s[0]:
                    fun.args = []
                else:
                    a = [self._parse_ref(arg.strip()) for arg in args.strip("()").split(",")]
                    assert all([isinstance(arg, tIndex) for arg in a]), "Expected a type reference in args!"
                    fun.args = a  # type: ignore
                typ = Type()
                typ.kind.value = 10  # Fun
                typ.definition = m_def
                code.types.append(typ)

    def _parse_ref(self, val: str) -> ResolvableVarInt:
        if val[1] != "@":
            raise SyntaxError("Expected a reference!")
        match val[0]:  # TODO: float, field support
            case "f":
                return fIndex(int(val[2:]))
            case "t":
                return tIndex(int(val[2:]))
            case "s":
                return strRef(int(val[2:]))
            case "g":
                return gIndex(int(val[2:]))
            case "i":
                return intRef(int(val[2:]))
            case "b":
                return bytesRef(int(val[2:]))
        raise SyntaxError(f"Unknown prefix '{val[0]}'!")

    def _parse_opcode_ref(self, val: str) -> Any:
        if val[1] != "@":
            if val[0] == '"':
                return self._get_str_idx(val[1:-1])
            elif val.startswith("reg"):
                return Reg(int(val[3:]))
        match val[0]:  # TODO: float, field support
            case "f":
                return fIndex(int(val[2:]))
            case "t":
                return tIndex(int(val[2:]))
            case "s":
                return strRef(int(val[2:]))
            case "g":
                return gIndex(int(val[2:]))
            case "i":
                return intRef(int(val[2:]))
            case "b":
                return bytesRef(int(val[2:]))
        raise SyntaxError(f"Unknown prefix '{val[0]}'!")

    def _get_single_val(self, name: str) -> str:
        if len(self.raw_sections[name].value) != 1:
            raise SyntaxError(f"Expected exactly one value for '{name}'!")
        val = self.raw_sections[name].value[0]
        if isinstance(val, AsmValueStr):
            return val.value
        raise SyntaxError(f"Expected a string value for '{name}'!")

    def _get_str_idx(self, val: str) -> strRef:
        if val not in self.strings:
            self.strings.append(val)
        return strRef(self.strings.index(val))

    def _validate(self, code: Bytecode) -> None:
        if not code.entrypoint:
            raise SyntaxError("No entrypoint specified!")
        if not code.types:
            raise SyntaxError("No types specified!")
        code.entrypoint.resolve(code)

    def _add_natives(self, code: Bytecode, section: AsmSection) -> None:
        for n in section.value:
            if not isinstance(n, AsmValueStr):
                continue
            parts = n.value.split()
            assert len(parts) == 3, "Incorrect native structure!"
            assert parts[1].startswith("("), f"Unexpected token {parts[1][0]}"
            idx, typ, name = parts
            _idx = self._parse_ref(idx)
            assert isinstance(_idx, fIndex), "Native index must be a function reference!"
            _typ = self._parse_ref(typ.strip("()"))
            assert isinstance(_typ, tIndex), "Native Fun type must be a type reference!"
            lib, name = name.split(".")
            _lib = self._get_str_idx(lib)
            _name = self._get_str_idx(name)
            obj = Native()
            obj.findex = _idx
            obj.lib = _lib
            obj.name = _name
            obj.type = _typ
            code.natives.append(obj)

    def _opcode(self, val: str) -> Opcode:
        def remove_commas_outside_quotes(text: str) -> str:
            result = ""
            in_quotes = False
            for char in text:
                if char == '"':
                    in_quotes = not in_quotes
                if char == "," and not in_quotes:
                    result += " "
                else:
                    result += char
            return result

        val = remove_commas_outside_quotes(val)

        parts = re.findall(r"\"[^\"]*\"|\S+", val)
        assert len(parts) >= 1, "Opcode must have at least one part!"

        op = Opcode()
        name = parts[0]
        assert name in opcodes, f"Unknown opcode '{name}'!"
        op.op = name
        op.df = {}

        for i, (k, v) in enumerate(opcodes[name].items()):
            if i + 1 >= len(parts):
                raise SyntaxError(f"Not enough arguments for opcode {name}, expected {k}")
            typ = Opcode.TYPE_MAP[v]
            parsed = self._parse_opcode_ref(parts[i + 1])
            assert isinstance(parsed, typ), f"Expected type {typ} for argument {k} of opcode {name}, got {type(parsed)}"
            op.df[k] = parsed

        return op

    def _add_functions(self, code: Bytecode) -> None:
        for section in self.raw_sections.values():
            if section.name.startswith("f@"):
                func = Function()
                returns_section = section.get("returns")
                if isinstance(returns_section.value[0], AsmValueStr):
                    typ = self._parse_ref(returns_section.value[0].value)
                else:
                    raise SyntaxError("Return type must be a string reference!")

                assert isinstance(typ, tIndex), "Return type must be a type reference!"
                func.type = typ
                findex = self._parse_ref(section.name)
                assert isinstance(findex, fIndex), "Function index must be a function reference!"
                func.findex = findex

                regs_section = section.get("regs")
                regs: List[tIndex] = []
                for reg in regs_section.value:
                    if isinstance(reg, AsmValueStr):
                        res = self._parse_ref(reg.value)
                        assert isinstance(res, tIndex), "Register must be a type index!"
                        regs.append(res)
                    else:
                        raise SyntaxError("Register must be a string reference!")

                assert all(isinstance(r, tIndex) for r in regs), "All registers must be types!"
                func.regs = regs

                ops_section = section.get("ops")
                ops = []
                for op in ops_section.value:
                    if isinstance(op, AsmValueStr):
                        ops.append(self._opcode(op.value))
                    else:
                        raise SyntaxError("Operation must be a string!")

                func.ops = ops
                func.has_debug = False
                func.version = code.version.value
                code.functions.append(func)

    def _add_strings(self, code: Bytecode) -> None:
        for s in self.strings:
            code.strings.value.append(s)

    def assemble(self) -> Bytecode:
        required = ["version", "types", "entrypoint"]
        for req in required:
            assert req in self.raw_sections
        code = Bytecode.create_empty(
            no_extra_types=True,
            version=int(self._get_single_val("version")),
        )
        self._add_types(code, self.raw_sections["types"])
        e = self._parse_ref(self._get_single_val("entrypoint"))
        assert isinstance(e, fIndex), "Entrypoint must be a function reference!"
        code.entrypoint = e
        if "natives" in self.raw_sections:
            self._add_natives(code, self.raw_sections["natives"])
        self._add_functions(code)
        self._add_strings(code)
        self._validate(code)
        return code


__all__ = ["AsmValue", "AsmValueStr", "AsmFile", "AsmSection"]
